from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.orchestration.db import safe_db_query
from mage_ai.orchestration.db.models.schedules import BlockRun


class OutputResource(GenericResource):
    @classmethod
    @safe_db_query
    def collection(self, query, meta, user, **kwargs):
        parent_model = kwargs['parent_model']

        outputs = []
        if type(parent_model) is BlockRun:
            outputs = parent_model.get_outputs()

        return self.build_result_set(
            outputs,
            user,
            **kwargs,
        )



import os
from unittest.mock import patch

import pandas as pd
from async_timeout import asyncio
from faker import Faker
from pandas.testing import assert_frame_equal

# from mage_ai.data_cleaner.column_types.constants import ColumnType
from mage_ai.data_preparation.models.block import Block, BlockType
from mage_ai.data_preparation.models.block.errors import HasDownstreamDependencies
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.data_preparation.repo_manager import get_repo_config
from mage_ai.data_preparation.variable_manager import VariableManager
from mage_ai.tests.base_test import DBTestCase


class BlockTest(DBTestCase):
    def test_create(self):
        block1 = Block.create('test_transformer', 'transformer', self.repo_path)
        block2 = Block.create('test data loader', BlockType.DATA_LOADER, self.repo_path)
        self.assertTrue(os.path.exists(f'{self.repo_path}/transformers/test_transformer.py'))
        self.assertTrue(os.path.exists(f'{self.repo_path}/transformers/__init__.py'))
        self.assertTrue(os.path.exists(f'{self.repo_path}/data_loaders/test_data_loader.py'))
        self.assertTrue(os.path.exists(f'{self.repo_path}/data_loaders/__init__.py'))
        self.assertEqual(block1.name, 'test_transformer')
        self.assertEqual(block1.uuid, 'test_transformer')
        self.assertEqual(block1.type, 'transformer')
        self.assertEqual(block2.name, 'test data loader')
        self.assertEqual(block2.uuid, 'test_data_loader')
        self.assertEqual(block2.type, 'data_loader')

    def test_execute(self):
        pipeline = Pipeline.create(
            'test pipeline',
            repo_path=self.repo_path,
        )
        block1 = Block.create('test_data_loader', 'data_loader', self.repo_path, pipeline=pipeline)
        block2 = Block.create(
            'test_transformer',
            'transformer',
            self.repo_path,
            pipeline=pipeline,
            upstream_block_uuids=['test_data_loader'],
        )
        with open(block1.file_path, 'w') as file:
            file.write('''import pandas as pd
@data_loader
def load_data():
    data = {'col1': [1, 1, 3], 'col2': [2, 2, 4]}
    df = pd.DataFrame(data)
    return [df]
            ''')
        with open(block2.file_path, 'w') as file:
            file.write('''import pandas as pd
@transformer
def remove_duplicate_rows(df):
    df_transformed = df.drop_duplicates()
    return [df_transformed]
            ''')
        asyncio.run(block1.execute(analyze_outputs=True))
        asyncio.run(block2.execute(analyze_outputs=True))

        variable_manager = VariableManager(
            variables_dir=get_repo_config(self.repo_path).variables_dir,
        )
        data = variable_manager.get_variable(
            pipeline.uuid,
            block2.uuid,
            'output_0',
            variable_type='dataframe'
        )
        analysis = variable_manager.get_variable(
            pipeline.uuid,
            block2.uuid,
            'output_0',
            variable_type='dataframe_analysis',
        )
        df_final = pd.DataFrame({'col1': [1, 1, 3], 'col2': [2, 2, 4]}).drop_duplicates()
        assert_frame_equal(data, df_final)

        analysis
        # TODO (Xiaoyou Wang): uncomment this one serialization of block output is fixed.
        # self.assertEqual(
        #     analysis['metadata']['column_types'],
        #     dict(col1=ColumnType.TRUE_OR_FALSE, col2=ColumnType.TRUE_OR_FALSE),
        # )
        # self.assertTrue(len(analysis['statistics']) > 0)
        # self.assertTrue(len(analysis['insights']) > 0)

    def test_execute_dicts_and_lists(self):
        pipeline = Pipeline.create(
            'test_pipeline_execute_dicts_and_lists',
            repo_path=self.repo_path,
        )
        block1 = Block.create(
            'test_data_loader_2',
            'data_loader',
            self.repo_path,
            pipeline=pipeline,
        )
        block2 = Block.create(
            'test_transformer_2',
            'transformer',
            self.repo_path,
            pipeline=pipeline,
            upstream_block_uuids=['test_data_loader_2'],
        )
        with open(block1.file_path, 'w') as file:
            file.write('''import pandas as pd
@data_loader
def load_data():
    data = {
        'col1': [1, 1, 3],
        'col2': [2, 2, 4],
        'col3': [dict(mage=1), dict(mage=2), dict(mage=3)],
        'col4': [[dict(mage=1)], [dict(mage=2)], [dict(mage=3)]],
    }
    df = pd.DataFrame(data)
    return [df]
            ''')
        with open(block2.file_path, 'w') as file:
            file.write('''import pandas as pd
@transformer
def remove_duplicate_rows(df):
    df_transformed = df
    return [df_transformed]
            ''')
        asyncio.run(block1.execute())
        asyncio.run(block2.execute())

        variable_manager = VariableManager(
            variables_dir=get_repo_config(self.repo_path).variables_dir,
        )
        data = variable_manager.get_variable(
            pipeline.uuid,
            block2.uuid,
            'output_0',
            variable_type='dataframe'
        )
        df_final = pd.DataFrame({
            'col1': [1, 1, 3],
            'col2': [2, 2, 4],
            'col3': [dict(mage=1), dict(mage=2), dict(mage=3)],
            'col4': [[dict(mage=1)], [dict(mage=2)], [dict(mage=3)]],
        })
        assert_frame_equal(data, df_final)

    def test_execute_multiple_upstream_blocks(self):
        pipeline = Pipeline.create(
            'test pipeline 2',
            repo_path=self.repo_path,
        )
        block1 = Block.create(
            'test_data_loader_1',
            'data_loader',
            self.repo_path,
            pipeline=pipeline,
        )
        block2 = Block.create(
            'test_data_loader_2',
            'data_loader',
            self.repo_path,
            pipeline=pipeline,
        )
        block3 = Block.create(
            'test_transformer',
            'transformer',
            self.repo_path,
            pipeline=pipeline,
            upstream_block_uuids=['test_data_loader_1', 'test_data_loader_2'],
        )
        with open(block1.file_path, 'w') as file:
            file.write('''import pandas as pd
@data_loader
def load_data():
    data = {'col1': [1, 3], 'col2': [2, 4]}
    df = pd.DataFrame(data)
    return [df]
            ''')
        with open(block2.file_path, 'w') as file:
            file.write('''import pandas as pd
@data_loader
def load_data():
    data = {'col1': [5], 'col2': [6]}
    df = pd.DataFrame(data)
    return [df]
            ''')
        with open(block3.file_path, 'w') as file:
            file.write('''import pandas as pd
@transformer
def union_datasets(df1, df2):
    df_union = pd.concat([df1, df2]).reset_index(drop=True)
    return [df_union]
            ''')
        asyncio.run(block1.execute(analyze_outputs=True))
        asyncio.run(block2.execute(analyze_outputs=True))
        asyncio.run(block3.execute(analyze_outputs=True))

        variable_manager = VariableManager(
            variables_dir=get_repo_config(self.repo_path).variables_dir,
        )
        data = variable_manager.get_variable(
            pipeline.uuid,
            block3.uuid,
            'output_0',
            variable_type='dataframe'
        )
        analysis = variable_manager.get_variable(
            pipeline.uuid,
            block3.uuid,
            'output_0',
            variable_type='dataframe_analysis',
        )
        df_final = pd.concat([
            pd.DataFrame({'col1': [1, 3], 'col2': [2, 4]}),
            pd.DataFrame({'col1': [5], 'col2': [6]}),
        ]).reset_index(drop=True)

        assert_frame_equal(data, df_final)

        analysis
        # TODO (Xiaoyou Wang): uncomment this one serialization of block output is fixed.
        # self.assertEqual(
        #     analysis['metadata']['column_types'],
        #     dict(col1=ColumnType.NUMBER, col2=ColumnType.NUMBER),
        # )
        # self.assertTrue(len(analysis['statistics']) > 0)
        # self.assertTrue(len(analysis['insights']) > 0)
        # self.assertTrue(len(analysis['suggestions']) == 0)

    def test_execute_validation(self):
        pipeline = Pipeline.create(
            'test pipeline 3',
            repo_path=self.repo_path,
        )
        block1 = Block.create(
            'test_data_loader_1',
            'data_loader',
            self.repo_path,
            pipeline=pipeline,
        )
        block2 = Block.create(
            'test_data_loader_2',
            'data_loader',
            self.repo_path,
            pipeline=pipeline,
        )
        block3 = Block.create(
            'test_transformer',
            'transformer',
            self.repo_path,
            pipeline=pipeline,
            upstream_block_uuids=['test_data_loader_1', 'test_data_loader_2'],
        )
        with open(block1.file_path, 'w') as file:
            file.write('''import pandas as pd
@data_loader
def load_data():
    data = {'col1': [1, 3], 'col2': [2, 4]}
    df = pd.DataFrame(data)
    return [df]
            ''')
        with open(block2.file_path, 'w') as file:
            file.write('''import pandas as pd
@data_loader
def load_data():
    data = {'col1': [5], 'col2': [6]}
    df = pd.DataFrame(data)
    return [df]
            ''')
        with open(block3.file_path, 'w') as file:
            file.write('''import pandas as pd
@transformer
def incorrect_function(df1):
    return df1
            ''')
        asyncio.run(block1.execute())
        asyncio.run(block2.execute())
        with self.assertRaises(Exception):
            asyncio.run(block3.execute())

        with open(block3.file_path, 'w') as file:
            file.write('''import pandas as pd
@transformer
def incorrect_function(df1, df2, df3):
    return df1
            ''')
        with self.assertRaises(Exception):
            asyncio.run(block3.execute())

    def test_sensor_block_args_execution(self):
        pipeline = Pipeline.create(
            'test pipeline 7',
            repo_path=self.repo_path,
        )
        block1 = Block.create(
            'test_data_loader_1',
            'data_loader',
            self.repo_path,
            pipeline=pipeline,
        )
        block2 = Block.create(
            'test_sensor',
            'sensor',
            self.repo_path,
            pipeline=pipeline,
        )
        block2.upstream_blocks = [block1]
        block1.downstream_blocks = [block2]
        with open(block1.file_path, 'w') as file:
            file.write('''import pandas as pd
@data_loader
def load_data():
    data = {'col1': [1, 3], 'col2': [2, 4]}
    df = pd.DataFrame(data)
    return [df]
            ''')
        with open(block2.file_path, 'w') as file:
            file.write('''import pandas as pd
@sensor
def sensor(*args):
    df = args[0]
    return df is not None
            ''')
        asyncio.run(block1.execute())
        output = block2.execute_sync(test_execution=True)
        self.assertEqual(output['output'][0], True)

    @patch('builtins.print')
    def test_execute_with_callback_success(self, mock_print):
        pipeline = Pipeline.create(
            'test pipeline 4',
            repo_path=self.repo_path,
        )
        block1 = Block.create(
            'test_data_loader_1',
            'data_loader',
            self.repo_path,
            pipeline=pipeline,
        )
        with open(block1.file_path, 'w') as file:
            file.write('''import pandas as pd
@data_loader
def load_data():
    data = {'col1': [1, 3], 'col2': [2, 4]}
    df = pd.DataFrame(data)
    return [df]
            ''')
        block1.update(dict(has_callback=True))
        with open(block1.callback_block.file_path, 'w') as file:
            file.write('''
@on_success
def on_success_callback(**kwargs):
    print('SUCCESS')
            ''')
        block1.execute_with_callback()
        mock_print.assert_called_with('SUCCESS')

    @patch('builtins.print')
    def test_execute_with_callback_failure(self, mock_print):
        pipeline = Pipeline.create(
            'test pipeline 5',
            repo_path=self.repo_path,
        )
        block1 = Block.create(
            'test_data_loader_1',
            'data_loader',
            self.repo_path,
            pipeline=pipeline,
        )
        with open(block1.file_path, 'w') as file:
            file.write('''import pandas as pd
@data_loader
def load_data():
    raise Exception('failed')
            ''')
        block1.update(dict(has_callback=True))
        with open(block1.callback_block.file_path, 'w') as file:
            file.write('''
@on_failure
def on_failure_callback(**kwargs):
    print('FAILED')
            ''')

        with self.assertRaises(Exception):
            block1.execute_with_callback()
            mock_print.assert_called_with('FAILED')

    def test_to_dict(self):
        block1 = Block.create(
            'test_transformer_2',
            'transformer',
            self.repo_path,
            language='sql',
        )
        block2 = Block.create(
            'test_data_exporter',
            'data_exporter',
            self.repo_path,
            language='python',
        )
        block2.upstream_blocks = [block1]
        block1.downstream_blocks = [block2]
        self.assertEqual(block1.to_dict(), dict(
            all_upstream_blocks_executed=True,
            color=None,
            configuration={},
            downstream_blocks=['test_data_exporter'],
            executor_config=None,
            executor_type='local_python',
            has_callback=False,
            language='sql',
            name='test_transformer_2',
            retry_config=None,
            status='not_executed',
            type='transformer',
            upstream_blocks=[],
            uuid='test_transformer_2',
        ))
        self.assertEqual(block2.to_dict(), dict(
            all_upstream_blocks_executed=False,
            color=None,
            configuration={},
            downstream_blocks=[],
            executor_config=None,
            executor_type='local_python',
            has_callback=False,
            language='python',
            name='test_data_exporter',
            retry_config=None,
            status='not_executed',
            type='data_exporter',
            upstream_blocks=['test_transformer_2'],
            uuid='test_data_exporter',
        ))

    def test_full_table_name(self):
        faker = Faker()
        pipeline = Pipeline.create(faker.name(), repo_path=self.repo_path)
        block = Block.create(faker.name(), 'data_loader', self.repo_path, pipeline=pipeline)

        def get_block():
            return Pipeline.get(pipeline.uuid).get_block(block.uuid)

        with open(block.file_path, 'w') as file:
            file.write('')
        self.assertEqual(get_block().full_table_name, None)

        with open(block.file_path, 'w') as file:
            file.write("""
CREATE TABLE mage.users_v1 (
    id BIGINT
);
""")
        self.assertEqual(get_block().full_table_name, 'mage.users_v1')

        with open(block.file_path, 'w') as file:
            file.write("""
CREATE TABLE mage.users_v1 (
    id BIGINT
);

INSERT INTO mage.users_v2
SELECT 1 AS id;
""")
        self.assertEqual(get_block().full_table_name, 'mage.users_v1')

        with open(block.file_path, 'w') as file:
            file.write("""
INSERT INTO mage.users_v2
SELECT 1 AS id;
""")
        self.assertEqual(get_block().full_table_name, 'mage.users_v2')

        with open(block.file_path, 'w') as file:
            file.write("""
INSERT INTO mage.users_v2
SELECT 1 AS id;

INSERT INTO mage.users_v3
SELECT 1 AS id;
""")
        self.assertEqual(get_block().full_table_name, 'mage.users_v3')

        with open(block.file_path, 'w') as file:
            file.write("""
INSERT mage.users_v2
SELECT 1 AS id;
""")
        self.assertEqual(get_block().full_table_name, 'mage.users_v2')

        with open(block.file_path, 'w') as file:
            file.write("""
INSERT OVERWRITE INTO mage.users_v2
SELECT 1 AS id;
""")
        self.assertEqual(get_block().full_table_name, 'mage.users_v2')

    def test_delete(self):
        pipeline = Pipeline.create(
            'test pipeline delete',
            repo_path=self.repo_path,
        )
        block1 = Block.create('test_data_loader', 'data_loader', self.repo_path, pipeline=pipeline)
        block2 = Block.create(
            'test_transformer',
            'transformer',
            self.repo_path,
            pipeline=pipeline,
            upstream_block_uuids=['test_data_loader'],
        )

        self.assertRaises(HasDownstreamDependencies, block1.delete)

        block2.delete()
        self.assertIsNone(pipeline.get_block('test_transformer'))

    def test_delete_force(self):
        pipeline = Pipeline.create(
            'test_pipeline_delete_force',
            repo_path=self.repo_path,
        )
        block1 = Block.create('test_data_loader', 'data_loader', self.repo_path, pipeline=pipeline)
        Block.create(
            'test_transformer',
            'transformer',
            self.repo_path,
            pipeline=pipeline,
            upstream_block_uuids=['test_data_loader'],
        )

        test_block = Block.get_block('test_data_loader', 'test_data_loader', 'data_loader')

        self.assertRaises(HasDownstreamDependencies, block1.delete)
        self.assertRaises(HasDownstreamDependencies, test_block.delete)

        test_block.delete(force=True)

        pipeline = Pipeline.get('test_pipeline_delete_force')
        self.assertIsNone(pipeline.get_block('test_data_loader'))

        block2 = pipeline.get_block('test_transformer')
        self.assertEqual([], block2.upstream_block_uuids)



from mage_ai.api.operations import constants
from mage_ai.api.presenters.BasePresenter import BasePresenter
from mage_ai.data_preparation.models.block.dbt.utils import (
    add_blocks_upstream_from_refs,
    compiled_query_string,
)


class BlockPresenter(BasePresenter):
    default_attributes = [
        'all_upstream_blocks_executed',
        'callback_blocks',
        'color',
        'conditional_blocks',
        'configuration',
        'downstream_blocks',
        'executor_config',
        'executor_type',
        'has_callback',
        'language',
        'metadata',
        'name',
        'replicated_block',
        'retry_config',
        'status',
        'type',
        'upstream_blocks',
        'uuid',
    ]

    def present(self, **kwargs):
        display_format = kwargs['format']

        if display_format in [constants.CREATE, constants.UPDATE]:
            return self.model.to_dict(include_content=True)
        elif display_format in [constants.DETAIL, 'dbt']:
            query = kwargs.get('query', {})

            include_outputs = query.get('include_outputs', [True])
            if include_outputs:
                include_outputs = include_outputs[0]

            state_stream = query.get('state_stream', [None])
            if state_stream:
                state_stream = state_stream[0]
            destination_table = query.get('destination_table', [None])
            if destination_table:
                destination_table = destination_table[0]

            data = self.model.to_dict(
                destination_table=destination_table,
                include_content=True,
                include_outputs=include_outputs,
                state_stream=state_stream,
            )

            if 'dbt' == display_format:
                upstream_blocks = add_blocks_upstream_from_refs(
                    self.model,
                    add_current_block=True,
                    read_only=True,
                )
                query_string = compiled_query_string(self.model)
                data['metadata'] = dict(dbt=dict(
                    lineage=[b.to_dict() for b in upstream_blocks],
                    sql=query_string,
                ))

            return data

        return self.model.to_dict()


BlockPresenter.register_formats([
    constants.CREATE,
    constants.UPDATE,
], [
    'content',
])

BlockPresenter.register_format(
    constants.DETAIL,
    [
        'content',
        'outputs',
    ],
)



from azure.identity import DefaultAzureCredential, ClientSecretCredential
from azure.storage.blob import BlobServiceClient
from io import BytesIO
from mage_ai.io.base import BaseFile, FileFormat, QUERY_ROW_LIMIT
from mage_ai.io.config import BaseConfigLoader, ConfigKey
from mage_ai.shared.hash import extract, merge_dict
from pandas import DataFrame
from typing import Union


class AzureBlobStorage(BaseFile):
    """
    Handles data transfer between a Azure Blob Storage container and the Mage app. Supports
    loading files of any of the following types:
    - ".csv"
    - ".json"
    - ".parquet"
    """
    AUTHENTICATION_KEYS = [
        ConfigKey.AZURE_CLIENT_ID,
        ConfigKey.AZURE_CLIENT_SECRET,
        ConfigKey.AZURE_TENANT_ID,
    ]

    def __init__(
        self,
        verbose: bool = False,
        **kwargs,
    ) -> None:
        """
        Initializes data loader from an Azure Blob Storage container.
        """

        super().__init__(verbose=kwargs.get('verbose', True))

        if all(key in kwargs for key in self.AUTHENTICATION_KEYS):
            self.creds = ClientSecretCredential(
                kwargs[ConfigKey.AZURE_TENANT_ID],
                kwargs[ConfigKey.AZURE_CLIENT_ID],
                kwargs[ConfigKey.AZURE_CLIENT_SECRET],
            )
        else:
            self.creds = DefaultAzureCredential()
        self.service_client = BlobServiceClient(
            account_url=f'https://{kwargs["storage_account_name"]}.blob.core.windows.net',
            credential=self.creds,
        )

    def load(
        self,
        container_name: str,
        blob_path: str,
        format: Union[FileFormat, str, None] = None,
        limit: int = QUERY_ROW_LIMIT,
        **kwargs,
    ) -> DataFrame:
        """
        Loads data from Azure Blob Storage into a Pandas data frame. This function will load at
        maximum 10,000,000 rows of data from the specified file.

        Returns:
            DataFrame: The data frame constructed from the file in the Azure Blob Storage container.
        """
        if format is None:
            format = self._get_file_format(blob_path)
        with self.printer.print_msg(
            f'Loading data frame from container \'{container_name}\' at path \'{blob_path}\''
        ):
            blob_client = self.service_client.get_blob_client(
                container_name,
                blob_path,
            )
            if not blob_client.exists():
                return DataFrame()
            blob_data = blob_client.download_blob()
            buffer = BytesIO(blob_data.readall())
            return self._read(buffer, format, limit, **kwargs)

    def export(
        self,
        df: DataFrame,
        container_name: str,
        blob_path: str,
        format: Union[FileFormat, str, None] = None,
        **kwargs,
    ) -> None:
        """
        Exports data frame to an Azure Blob Storage container.
        """
        if format is None:
            format = self._get_file_format(blob_path)

        with self.printer.print_msg(
            f'Exporting data frame to container \'{container_name}\' at path \'{blob_path}\''
        ):
            blob_client = self.service_client.get_blob_client(
                container_name,
                blob_path,
            )
            buffer = BytesIO()
            self._write(df, format, buffer, **kwargs)
            buffer.seek(0)
            blob_client.upload_blob(buffer.read())

    def exists(
        self,
        container_name: str,
        prefix: str,
    ) -> bool:
        """
        Checks if content exists at a certain path in an Azure Blob Storage container.
        """
        blob_client = self.service_client.get_blob_client(
            container_name,
            prefix,
        )
        return blob_client.exists()

    @classmethod
    def with_config(
        cls,
        config: BaseConfigLoader,
        **kwargs,
    ) -> 'AzureBlobStorage':
        """
        Initializes Azure Blob Storage client from configuration loader

        Args:
            config (BaseConfigLoader): Configuration loader object
        """
        if ConfigKey.AZURE_STORAGE_ACCOUNT_NAME not in config:
            raise ValueError(
                'Require AZURE_STORAGE_ACCOUNT_NAME to be specified for Azure Blob Storage.'
            )

        return cls(
            storage_account_name=config[ConfigKey.AZURE_STORAGE_ACCOUNT_NAME],
            **merge_dict(
                extract(config, cls.AUTHENTICATION_KEYS),
                kwargs,
            ),
        )



"""Add indexes to pipeline run.

Revision ID: 2266370f589b
Revises: 5cd59ec4cf1d
Create Date: 2022-11-01 18:21:53.877930

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '2266370f589b'
down_revision = '5cd59ec4cf1d'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_index(op.f('ix_pipeline_run_execution_date'), 'pipeline_run', ['execution_date'], unique=False)
    op.create_index(op.f('ix_pipeline_run_pipeline_uuid'), 'pipeline_run', ['pipeline_uuid'], unique=False)
    op.create_index(op.f('ix_pipeline_run_status'), 'pipeline_run', ['status'], unique=False)
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f('ix_pipeline_run_status'), table_name='pipeline_run')
    op.drop_index(op.f('ix_pipeline_run_pipeline_uuid'), table_name='pipeline_run')
    op.drop_index(op.f('ix_pipeline_run_execution_date'), table_name='pipeline_run')
    # ### end Alembic commands ###



from mage_ai.api.oauth_scope import OauthScope
from mage_ai.api.operations import constants
from mage_ai.api.policies.BasePolicy import BasePolicy
from mage_ai.api.presenters.BlockOutputPresenter import BlockOutputPresenter


class BlockOutputPolicy(BasePolicy):
    pass


BlockOutputPolicy.allow_actions([
    constants.DETAIL,
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], condition=lambda policy: policy.has_at_least_viewer_role())

BlockOutputPolicy.allow_read(BlockOutputPresenter.default_attributes, scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.DETAIL,
], condition=lambda policy: policy.has_at_least_viewer_role())

BlockOutputPolicy.allow_query([
    'pipeline_uuid',
], scopes=[
    OauthScope.CLIENT_PRIVATE,
], on_action=[
    constants.DETAIL,
], condition=lambda policy: policy.has_at_least_viewer_role())



from dataclasses import dataclass, field
from mage_ai.shared.config import BaseConfig
from typing import Dict, List


@dataclass
class SparkConfig(BaseConfig):
    app_name: str = None
    spark_master: str = None
    spark_home: str = None
    executor_env: Dict = field(default_factory=dict)
    spark_jars: List = field(default_factory=list)
    others: Dict = field(default_factory=dict)



from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.orchestration.db import safe_db_query
from typing import Dict
import os
import shutil
import urllib.parse


def full_path(*args) -> str:
    return os.path.join(get_repo_path(), *list(filter(lambda x: x and len(x) >= 1, args)))


class FolderResource(GenericResource):
    @classmethod
    @safe_db_query
    def create(self, payload: Dict, user, **kwargs) -> 'FolderResource':
        path = full_path(payload.get('path'), payload.get('name'))
        os.makedirs(path, exist_ok=True if payload.get('overwrite', False) else False)
        return self(dict(path=path), user, **kwargs)

    @classmethod
    def member(self, pk, user, **kwargs):
        path = full_path(urllib.parse.unquote(pk))
        return self(dict(path=path), user, **kwargs)

    @safe_db_query
    def delete(self, **kwargs):
        return shutil.rmtree(self.path)

    @safe_db_query
    def update(self, payload, **kwargs):
        path = full_path(payload.get('path'), payload.get('name'))
        shutil.move(self.path, path)
        self.model = dict(path=path)
        return self



from mage_ai.api.resources.GenericResource import GenericResource
from mage_ai.data_preparation.models.project import Project
from mage_ai.data_preparation.repo_manager import get_repo_config
from mage_ai.orchestration.db import safe_db_query
from mage_ai.shared.hash import merge_dict
from mage_ai.usage_statistics.logger import UsageStatisticLogger
import uuid


async def build_project(repo_config=None, **kwargs):
    project = Project(repo_config=repo_config)

    model = merge_dict(project.repo_config.to_dict(), dict(
        latest_version=await project.latest_version(),
        name=project.name,
        project_uuid=project.project_uuid,
        version=project.version,
    ))

    if kwargs:
        model.update(kwargs)

    return model


class ProjectResource(GenericResource):
    @classmethod
    @safe_db_query
    async def collection(self, query, meta, user, **kwargs):
        project = await self.member(None, user, **kwargs)
        collection = [project.model]

        return self.build_result_set(
            collection,
            user,
            **kwargs,
        )

    @classmethod
    @safe_db_query
    async def member(self, _, user, **kwargs):
        model = await build_project()
        return self(model, user, **kwargs)

    @safe_db_query
    async def update(self, payload, **kwargs):
        repo_config = get_repo_config()

        data = {}
        should_log_project = False

        if 'help_improve_mage' in payload:
            if payload['help_improve_mage']:
                should_log_project = True

                if not repo_config.project_uuid:
                    data['project_uuid'] = uuid.uuid4().hex

            data['help_improve_mage'] = payload['help_improve_mage']

        if len(data.keys()) >= 1:
            repo_config.save(**data)

        self.model = await build_project(repo_config)

        if should_log_project:
            await UsageStatisticLogger().project_impression()

        return self



import os
from datetime import datetime
from typing import Dict, List, Tuple

import aiofiles

from mage_ai.data_preparation.models.errors import FileExistsError
from mage_ai.data_preparation.repo_manager import get_repo_path

FILE_VERSIONS_DIR = '.file_versions'
BLACKLISTED_DIRS = frozenset([
    'venv',
    'env',
    '.git',
    '.logs',
    '.variables',
    '.DS_Store',
    '__pycache__',
    FILE_VERSIONS_DIR,
])
INACCESSIBLE_DIRS = frozenset(['__pycache__'])
MAX_DEPTH = 30
MAX_NUMBER_OF_FILE_VERSIONS = int(os.getenv('MAX_NUMBER_OF_FILE_VERSIONS', 100) or 100)

PIPELINES_FOLDER_PREFIX = f'pipelines{os.sep}'


class File:
    def __init__(
        self,
        filename: str,
        dir_path: str,
        repo_path: str,
    ) -> None:
        self.filename = filename
        self.dir_path = dir_path
        self.repo_path = repo_path

    @property
    def file_path(self) -> str:
        return os.path.join(self.repo_path, self.dir_path, self.filename)

    @classmethod
    def file_exists(self, file_path: str) -> bool:
        return os.path.isfile(file_path)

    @classmethod
    def create_parent_directories(self, file_path: str) -> bool:
        will_create = not self.file_exists(file_path)
        if will_create:
            os.makedirs(os.path.dirname(file_path), exist_ok=True)
        return will_create

    @classmethod
    def create(
        self,
        filename,
        dir_path,
        content: str = None,
        repo_path: str = None,
        create_directories_if_not_exist: bool = True,
        file_version_only: bool = False,
        overwrite: bool = True,
    ):
        repo_path = repo_path or get_repo_path()
        file = File(filename, dir_path, repo_path)

        self.write(
            repo_path,
            dir_path,
            filename,
            content,
            create_directories_if_not_exist=create_directories_if_not_exist,
            file_version_only=file_version_only,
            overwrite=overwrite,
        )

        return file

    @classmethod
    async def create_async(
        self,
        filename,
        dir_path,
        content: str = None,
        repo_path: str = None,
        create_directories_if_not_exist: bool = True,
        file_version_only: bool = False,
        overwrite: bool = True,
    ):
        repo_path = repo_path or get_repo_path()
        file = File(filename, dir_path, repo_path)

        await self.write_async(
            repo_path,
            dir_path,
            filename,
            content,
            create_directories_if_not_exist=create_directories_if_not_exist,
            file_version_only=file_version_only,
            overwrite=overwrite,
        )

        return file

    @classmethod
    def from_path(self, file_path, repo_path=None):
        repo_path = repo_path or get_repo_path()
        return File(os.path.basename(file_path), os.path.dirname(file_path), repo_path)

    @classmethod
    def get_all_files(self, repo_path):
        return traverse(os.path.basename(repo_path), True, repo_path)

    @classmethod
    def file_path_versions_dir(
        self,
        repo_path: str,
        dir_path: str,
        filename: str,
    ) -> Tuple[str, str]:
        return os.path.join(
            repo_path,
            FILE_VERSIONS_DIR,
            dir_path.replace(repo_path, os.path.join(repo_path, FILE_VERSIONS_DIR)),
            filename,
        )

    @classmethod
    def validate_content(self, dir_path, filename, content):
        if dir_path.startswith(PIPELINES_FOLDER_PREFIX) and filename == 'triggers.yaml':
            from mage_ai.data_preparation.models.triggers import load_trigger_configs

            pipeline_uuid = dir_path[len(PIPELINES_FOLDER_PREFIX):]
            load_trigger_configs(content, pipeline_uuid=pipeline_uuid, raise_exception=True)

    @classmethod
    def write_preprocess(
        self,
        repo_path: str,
        dir_path: str,
        filename: str,
        content: str,
        create_directories_if_not_exist: bool = True,
        file_version_only: bool = False,
        overwrite: bool = True,
    ):
        file_path_main = os.path.join(repo_path, dir_path, filename)
        file_path_versions_dir = self.file_path_versions_dir(repo_path, dir_path, filename)
        file_path_versions = os.path.join(
            file_path_versions_dir,
            str(round(datetime.utcnow().timestamp())),
        )

        arr = []

        if not file_version_only:
            arr.append((file_path_main, overwrite, create_directories_if_not_exist))

        if MAX_NUMBER_OF_FILE_VERSIONS >= 1:
            arr.append((file_path_versions, True, True))

            file_versions = []
            step = 0
            for _, _, files in os.walk(file_path_versions_dir):
                if step >= 1:
                    continue
                file_versions += files
                step += 1
            number_of_file_versions = len(file_versions)

            if number_of_file_versions >= MAX_NUMBER_OF_FILE_VERSIONS:
                number_of_file_versions_to_delete = 1 + \
                    (number_of_file_versions - MAX_NUMBER_OF_FILE_VERSIONS)
                for fn in sorted(file_versions)[:number_of_file_versions_to_delete]:
                    fn_path = os.path.join(file_path_versions_dir, fn)
                    os.remove(fn_path)

        for tup in arr:
            file_path, should_overwrite, should_create_directories = tup
            if self.file_exists(file_path) and not should_overwrite:
                raise FileExistsError(f'File at {file_path} already exists.')

            if should_create_directories:
                self.create_parent_directories(file_path)

            write_type = 'wb' if content and type(content) is bytes else 'w'
            yield file_path, write_type

    @classmethod
    def write(
        self,
        repo_path: str,
        dir_path: str,
        filename: str,
        content: str,
        create_directories_if_not_exist: bool = True,
        file_version_only: bool = False,
        overwrite: bool = True,
    ) -> None:
        for file_path, write_type in self.write_preprocess(
            repo_path,
            dir_path,
            filename,
            content,
            create_directories_if_not_exist=create_directories_if_not_exist,
            file_version_only=file_version_only,
            overwrite=overwrite,
        ):
            kwargs = dict(
                mode=write_type,
            )
            if write_type != 'wb':
                kwargs['encoding'] = 'utf-8'
            with open(file_path, **kwargs) as f:
                if content:
                    f.write(content)
        self.validate_content(dir_path, filename, content)

    @classmethod
    async def write_async(
        self,
        repo_path: str,
        dir_path: str,
        filename: str,
        content: str,
        create_directories_if_not_exist: bool = True,
        file_version_only: bool = False,
        overwrite: bool = True,
    ) -> None:
        for file_path, write_type in self.write_preprocess(
            repo_path,
            dir_path,
            filename,
            content,
            create_directories_if_not_exist=create_directories_if_not_exist,
            file_version_only=file_version_only,
            overwrite=overwrite,
        ):
            kwargs = dict(
                mode=write_type,
            )
            if write_type != 'wb':
                kwargs['encoding'] = 'utf-8'
            async with aiofiles.open(file_path, **kwargs) as fp:
                await fp.write(content)

    def exists(self) -> bool:
        return self.file_exists(self.file_path)

    def content(self):
        try:
            with open(self.file_path, encoding='utf-8') as fp:
                file_content = fp.read()
            return file_content
        except FileNotFoundError as err:
            print('file.content')
            print(err)
        return ''

    async def content_async(self):
        try:
            async with aiofiles.open(self.file_path, mode='r', encoding='utf-8') as fp:
                file_content = await fp.read()
            return file_content
        except FileNotFoundError as err:
            print(err)
        return ''

    def file_versions(self) -> List[str]:
        file_path_versions_dir = self.file_path_versions_dir(
            self.repo_path,
            self.dir_path,
            self.filename,
        )
        file_versions = []
        step = 0
        for _, _, files in os.walk(file_path_versions_dir):
            if step >= 1:
                continue
            file_versions += files
            step += 1

        file_path_versions_dir_without_repo = file_path_versions_dir.replace(
            os.path.join(self.repo_path, ''),
            '',
        )
        return [File(
            v,
            file_path_versions_dir_without_repo,
            self.repo_path,
        ) for v in sorted(file_versions, reverse=True)]

    def update_content(self, content: str):
        self.write(
            self.repo_path,
            self.dir_path,
            self.filename,
            content,
        )

    async def update_content_async(self, content: str):
        await self.write_async(
            self.repo_path,
            self.dir_path,
            self.filename,
            content,
        )

    def delete(self):
        os.remove(self.file_path)

    def rename(self, dir_path: str, filename):
        full_path = os.path.join(self.repo_path, dir_path, filename)
        self.create_parent_directories(full_path)
        os.rename(self.file_path, full_path)

        file_path_versions_dir = self.file_path_versions_dir(
            self.repo_path,
            self.dir_path,
            self.filename,
        )
        if os.path.exists(file_path_versions_dir):
            new_path = self.file_path_versions_dir(
                self.repo_path,
                dir_path,
                filename,
            )
            os.makedirs(new_path, exist_ok=True)
            os.rename(
                file_path_versions_dir,
                new_path,
            )

    def to_dict(self, include_content=False):
        data = dict(name=self.filename, path=os.path.join(self.dir_path, self.filename))
        if include_content:
            data['content'] = self.content()
        return data

    async def to_dict_async(self, include_content=False):
        data = dict(name=self.filename, path=os.path.join(self.dir_path, self.filename))
        if include_content:
            data['content'] = await self.content_async()
        return data


def traverse(name: str, is_dir: str, path: str, disabled=False, depth=1) -> Dict:
    tree_entry = dict(name=name)
    if not is_dir:
        tree_entry['disabled'] = disabled
        return tree_entry
    if depth >= MAX_DEPTH:
        return tree_entry
    can_access_children = name[0] == '.' or name in INACCESSIBLE_DIRS
    tree_entry['children'] = list(
        traverse(
            entry.name,
            entry.is_dir(follow_symlinks=False),
            entry.path,
            can_access_children,
            depth + 1,
        )
        for entry in sorted(
            filter(lambda entry: entry.name not in BLACKLISTED_DIRS, os.scandir(path)),
            key=lambda entry: entry.name,
        )
    )
    return tree_entry



from datetime import datetime as dt
from mage_ai.data_cleaner.transformer_actions.column import (
    add_column,
    count,
    count_distinct,
    clean_column_names,
    diff,
    # expand_column,
    first,
    fix_syntax_errors,
    last,
    remove_column,
    select,
    shift_down,
    shift_up,
)
from mage_ai.tests.base_test import TestCase
from pandas.testing import assert_frame_equal
from random import seed
import numpy as np
import pandas as pd

TEST_DATAFRAME = pd.DataFrame(
    [
        [1, 1000],
        [2, 1050],
        [1, 1100],
        [2, 1150],
    ],
    columns=[
        'group_id',
        'amount',
    ],
)


class ColumnTests(TestCase):
    def setUp(self):
        seed(42)
        return super().setUp()

    def test_remove_column(self):
        df = pd.DataFrame(
            [
                [0, False, 'a'],
                [1, True, 'b'],
            ],
            columns=[
                'integer',
                'boolean',
                'string',
            ],
        )

        action = dict(action_arguments=['string'])

        df_new = remove_column(df, action)
        self.assertEqual(
            df_new.to_dict(orient='records'),
            [
                dict(
                    integer=0,
                    boolean=False,
                ),
                dict(
                    integer=1,
                    boolean=True,
                ),
            ],
        )

        action = dict(action_arguments=['integer', 'boolean'])

        df_new = remove_column(df, action)
        self.assertEqual(
            df_new.to_dict(orient='records'),
            [
                dict(
                    string='a',
                ),
                dict(
                    string='b',
                ),
            ],
        )

    def test_add_column_addition(self):
        df = pd.DataFrame(
            [
                [1, 3, 7, 9],
                [4, 2, 9, 3],
            ],
            columns=[
                'integer1',
                'integer2',
                'integer3',
                'integer4',
            ],
        )
        action1 = dict(
            action_arguments=[
                'integer1',
                'integer2',
                'integer3',
            ],
            action_options={
                'udf': 'addition',
                'value': None,
            },
            outputs=[
                dict(
                    uuid='integer_addition',
                    column_type='number',
                ),
            ],
        )
        action2 = dict(
            action_arguments=['integer1'],
            action_options={
                'udf': 'addition',
                'value': 10,
            },
            outputs=[
                dict(
                    uuid='integer_addition2',
                    column_type='number',
                ),
            ],
        )
        action3 = dict(
            action_arguments=['integer1', 'integer4'],
            action_options={
                'udf': 'addition',
                'value': 10,
            },
            outputs=[
                dict(
                    uuid='integer_addition3',
                    column_type='number',
                ),
            ],
        )
        df_new = add_column(
            add_column(
                add_column(df, action1),
                action2,
            ),
            action3,
        )
        df_expected = pd.DataFrame(
            [
                [1, 3, 7, 9, 11, 11, 20],
                [4, 2, 9, 3, 15, 14, 17],
            ],
            columns=[
                'integer1',
                'integer2',
                'integer3',
                'integer4',
                'integer_addition',
                'integer_addition2',
                'integer_addition3',
            ],
        )
        assert_frame_equal(df_new, df_expected)

    def test_add_column_addition_days(self):
        df = pd.DataFrame(
            [
                ['2021-08-31'],
                ['2021-08-28'],
            ],
            columns=[
                'created_at',
            ],
        )
        action = dict(
            action_arguments=['created_at'],
            action_options=dict(
                column_type='datetime',
                time_unit='d',
                udf='addition',
                value=3,
            ),
            outputs=[
                dict(
                    uuid='3d_after_creation',
                    column_type='text',
                ),
            ],
        )
        df_new = add_column(df, action)
        df_expected = pd.DataFrame(
            [
                ['2021-08-31', '2021-09-03 00:00:00'],
                ['2021-08-28', '2021-08-31 00:00:00'],
            ],
            columns=['created_at', '3d_after_creation'],
        )
        assert_frame_equal(df_new, df_expected)

    def test_add_column_constant(self):
        df = pd.DataFrame(
            [
                [False],
                [True],
            ],
            columns=[
                'boolean',
            ],
        )
        action = dict(
            action_arguments=[10],
            action_options=dict(
                udf='constant',
            ),
            outputs=[
                dict(
                    uuid='integer',
                    column_type='number',
                ),
            ],
        )
        df_new = add_column(df, action)
        self.assertEqual(
            df_new.to_dict(orient='records'),
            [
                dict(
                    boolean=False,
                    integer=10,
                ),
                dict(
                    boolean=True,
                    integer=10,
                ),
            ],
        )

    def test_add_column_date_trunc(self):
        df = pd.DataFrame(
            [
                ['2021-08-31', False],
                ['2021-08-28', True],
            ],
            columns=[
                'created_at',
                'boolean',
            ],
        )
        action = dict(
            action_arguments=['created_at'],
            action_options=dict(
                udf='date_trunc',
                date_part='week',
            ),
            outputs=[
                dict(
                    uuid='week_date',
                    column_type='text',
                ),
            ],
        )
        df_new = add_column(df, action)
        self.assertEqual(
            df_new.to_dict(orient='records'),
            [
                dict(
                    created_at='2021-08-31',
                    boolean=False,
                    week_date='2021-08-30',
                ),
                dict(
                    created_at='2021-08-28',
                    boolean=True,
                    week_date='2021-08-23',
                ),
            ],
        )

    def test_add_column_difference(self):
        df = pd.DataFrame(
            [
                [1, 3],
                [4, 2],
            ],
            columns=[
                'integer1',
                'integer2',
            ],
        )
        action1 = dict(
            action_arguments=['integer1', 'integer2'],
            action_options={
                'udf': 'difference',
            },
            outputs=[
                dict(
                    uuid='integer_difference',
                    column_type='number',
                ),
            ],
        )
        action2 = dict(
            action_arguments=['integer1'],
            action_options={
                'udf': 'difference',
                'value': 10,
            },
            outputs=[
                dict(
                    uuid='integer_difference2',
                    column_type='number',
                ),
            ],
        )
        df_new = add_column(add_column(df, action1), action2)
        df_expected = pd.DataFrame(
            [
                [1, 3, -2, -9],
                [4, 2, 2, -6],
            ],
            columns=['integer1', 'integer2', 'integer_difference', 'integer_difference2'],
        )
        assert_frame_equal(df_new, df_expected)

    def test_add_column_difference_days(self):
        df = pd.DataFrame(
            [
                ['2021-08-31', '2021-09-14'],
                ['2021-08-28', '2021-09-03'],
            ],
            columns=[
                'created_at',
                'converted_at',
            ],
        )
        action = dict(
            action_arguments=['converted_at', 'created_at'],
            action_options=dict(
                column_type='datetime',
                time_unit='d',
                udf='difference',
            ),
            outputs=[
                dict(
                    uuid='days_diff',
                    column_type='number',
                ),
            ],
        )
        df_new = add_column(df, action)
        df_expected = pd.DataFrame(
            [
                ['2021-08-31', '2021-09-14', 14],
                ['2021-08-28', '2021-09-03', 6],
            ],
            columns=[
                'created_at',
                'converted_at',
                'days_diff',
            ],
        )
        assert_frame_equal(df_new, df_expected)

    def test_add_column_distance_between(self):
        df = pd.DataFrame(
            [
                [26.05308, -97.31838, 33.41939, -112.32606],
                [39.71954, -84.13056, 33.41939, -112.32606],
            ],
            columns=[
                'lat1',
                'lng1',
                'lat2',
                'lng2',
            ],
        )
        action = dict(
            action_arguments=['lat1', 'lng1', 'lat2', 'lng2'],
            action_options=dict(
                udf='distance_between',
            ),
            outputs=[
                dict(
                    uuid='distance',
                    column_type='number_with_decimals',
                ),
            ],
        )
        df_new = add_column(df, action)
        df_new['distance'] = df_new['distance'].round(3)
        self.assertEqual(
            df_new.to_dict(orient='records'),
            [
                dict(
                    lat1=26.05308,
                    lng1=-97.31838,
                    lat2=33.41939,
                    lng2=-112.32606,
                    distance=1661.898,
                ),
                dict(
                    lat1=39.71954,
                    lng1=-84.13056,
                    lat2=33.41939,
                    lng2=-112.32606,
                    distance=2601.545,
                ),
            ],
        )

    def test_add_column_divide(self):
        df = pd.DataFrame(
            [
                [12, 3, 70, 9],
                [4, 2, 90, 3],
            ],
            columns=[
                'integer1',
                'integer2',
                'integer3',
                'integer4',
            ],
        )
        action1 = dict(
            action_arguments=[
                'integer1',
                'integer2',
            ],
            action_options={
                'udf': 'divide',
            },
            outputs=[
                dict(
                    uuid='integer_divide',
                    column_type='number',
                ),
            ],
        )
        action2 = dict(
            action_arguments=['integer3'],
            action_options={
                'udf': 'divide',
                'value': 10,
            },
            outputs=[
                dict(
                    uuid='integer_divide2',
                    column_type='number',
                ),
            ],
        )
        df_new = add_column(add_column(df, action1), action2)
        df_expected = pd.DataFrame(
            [
                [12, 3, 70, 9, 4, 7],
                [4, 2, 90, 3, 2, 9],
            ],
            columns=[
                'integer1',
                'integer2',
                'integer3',
                'integer4',
                'integer_divide',
                'integer_divide2',
            ],
        )
        assert_frame_equal(df_new, df_expected)

    # def test_add_column_extract_dict_string(self):
    #     df = pd.DataFrame([
    #         '{\'country\': \'US\', \'age\': \'20\'}',
    #         '{\'country\': \'CA\'}',
    #         '{\'country\': \'UK\', \'age\': \'24\'}',
    #         '',
    #     ], columns=[
    #         'properties',
    #     ])
    #     action = dict(
    #         action_arguments=['properties', 'country'],
    #         action_options=dict(
    #             udf='extract_dict_value',
    #         ),
    #         outputs=[
    #             dict(
    #                 uuid='property_country',
    #                 column_type='text',
    #             ),
    #         ],
    #     )
    #     df_new = add_column(df, action)
    #     self.assertEqual(df_new.to_dict(orient='records'), [
    #         dict(
    #             properties='{\'country\': \'US\', \'age\': \'20\'}',
    #             property_country='US',
    #         ),
    #         dict(
    #             properties='{\'country\': \'CA\'}',
    #             property_country='CA',
    #         ),
    #         dict(
    #             properties='{\'country\': \'UK\', \'age\': \'24\'}',
    #             property_country='UK',
    #         ),
    #         dict(
    #             properties='',
    #             property_country=np.NaN,
    #         ),
    #     ])
    #     action2 = dict(
    #         action_arguments=['properties', 'age'],
    #         action_options=dict(
    #             udf='extract_dict_value',
    #         ),
    #         outputs=[
    #             dict(
    #                 uuid='property_age',
    #                 column_type='number',
    #             ),
    #         ],
    #     )
    #     df_new2 = add_column(df, action2)
    #     self.assertEqual(df_new2.to_dict(orient='records'), [
    #         dict(
    #             properties='{\'country\': \'US\', \'age\': \'20\'}',
    #             property_age=20,
    #         ),
    #         dict(
    #             properties='{\'country\': \'CA\'}',
    #             property_age=0,
    #         ),
    #         dict(
    #             properties='{\'country\': \'UK\', \'age\': \'24\'}',
    #             property_age=24,
    #         ),
    #         dict(
    #             properties='',
    #             property_age=0,
    #         ),
    #     ])

    # def test_add_column_extract_dict_string_with_json(self):
    #     df = pd.DataFrame([
    #         '{\"country\": \"US\", \"is_adult\": true}',
    #         '{\"country\": \"CA\"}',
    #         '{\"country\": \"UK\", \"is_adult\": false}',
    #         '',
    #     ], columns=[
    #         'properties',
    #     ])
    #     action = dict(
    #         action_arguments=['properties', 'country'],
    #         action_options=dict(
    #             udf='extract_dict_value',
    #         ),
    #         outputs=[
    #             dict(
    #                 uuid='property_country',
    #                 column_type='text',
    #             ),
    #         ],
    #     )
    #     df_new = add_column(df, action)
    #     self.assertEqual(df_new.to_dict(orient='records'), [
    #         dict(
    #             properties='{\"country\": \"US\", \"is_adult\": true}',
    #             property_country='US',
    #         ),
    #         dict(
    #             properties='{\"country\": \"CA\"}',
    #             property_country='CA',
    #         ),
    #         dict(
    #             properties='{\"country\": \"UK\", \"is_adult\": false}',
    #             property_country='UK',
    #         ),
    #         dict(
    #             properties='',
    #             property_country=np.NaN,
    #         ),
    #     ])
    #     action2 = dict(
    #         action_arguments=['properties', 'is_adult'],
    #         action_options=dict(
    #             udf='extract_dict_value',
    #         ),
    #         outputs=[
    #             dict(
    #                 uuid='property_is_adult',
    #                 column_type='true_or_false',
    #             ),
    #         ],
    #     )
    #     df_new2 = add_column(df, action2)
    #     self.assertEqual(df_new2.to_dict(orient='records'), [
    #         dict(
    #             properties='{\"country\": \"US\", \"is_adult\": true}',
    #             property_is_adult=True,
    #         ),
    #         dict(
    #             properties='{\"country\": \"CA\"}',
    #             property_is_adult=None,
    #         ),
    #         dict(
    #             properties='{\"country\": \"UK\", \"is_adult\": false}',
    #             property_is_adult=False,
    #         ),
    #         dict(
    #             properties='',
    #             property_is_adult=None,
    #         ),
    #     ])

    def test_add_column_formatted_date(self):
        df = pd.DataFrame(
            [
                ['2019-04-10 08:20:58', False],
                ['2019-03-05 03:30:30', True],
            ],
            columns=[
                'created_at',
                'boolean',
            ],
        )
        action = dict(
            action_arguments=['created_at'],
            action_options=dict(
                udf='formatted_date',
                format='%Y-%m-%d',
            ),
            outputs=[
                dict(
                    uuid='created_date',
                    column_type='text',
                ),
            ],
        )
        df_new = add_column(df, action)
        self.assertEqual(
            df_new.to_dict(orient='records'),
            [
                dict(
                    created_at='2019-04-10 08:20:58',
                    boolean=False,
                    created_date='2019-04-10',
                ),
                dict(
                    created_at='2019-03-05 03:30:30',
                    boolean=True,
                    created_date='2019-03-05',
                ),
            ],
        )

    def test_add_column_if_else(self):
        df = pd.DataFrame(
            [
                ['2019-04-10 08:20:58'],
                [None],
            ],
            columns=['converted_at'],
        )
        action = dict(
            action_arguments=[False, True],
            action_code='converted_at == null',
            action_options=dict(
                udf='if_else',
            ),
            outputs=[
                dict(
                    uuid='converted',
                    column_type='true_or_false',
                ),
            ],
        )
        df_new = add_column(df, action, original_df=df)
        self.assertEqual(
            df_new.to_dict(orient='records'),
            [
                dict(
                    converted_at='2019-04-10 08:20:58',
                    converted=True,
                ),
                dict(
                    converted_at=None,
                    converted=False,
                ),
            ],
        )

    def test_add_column_if_else_with_column(self):
        df = pd.DataFrame(
            [
                ['2019-04-10 08:20:58', 'test_user_id'],
                [None, None],
            ],
            columns=[
                'converted_at',
                'user_id',
            ],
        )
        action = dict(
            action_arguments=['unknown', 'user_id'],
            action_code='converted_at == null',
            action_options=dict(
                udf='if_else',
                arg1_type='value',
                arg2_type='column',
            ),
            outputs=[
                dict(
                    uuid='user_id_clean',
                    column_type='text',
                ),
            ],
        )
        df_new = add_column(df, action, original_df=df)
        self.assertEqual(
            df_new.to_dict(orient='records'),
            [
                dict(
                    converted_at='2019-04-10 08:20:58',
                    user_id='test_user_id',
                    user_id_clean='test_user_id',
                ),
                dict(
                    converted_at=None,
                    user_id=None,
                    user_id_clean='unknown',
                ),
            ],
        )

    def test_add_column_multiply(self):
        df = pd.DataFrame(
            [
                [1, 3, 7, 9],
                [4, 2, 9, 3],
            ],
            columns=[
                'integer1',
                'integer2',
                'integer3',
                'integer4',
            ],
        )
        action1 = dict(
            action_arguments=[
                'integer1',
                'integer2',
            ],
            action_options={
                'udf': 'multiply',
            },
            outputs=[
                dict(
                    uuid='integer_multiply',
                    column_type='number',
                ),
            ],
        )
        action2 = dict(
            action_arguments=['integer3'],
            action_options={
                'udf': 'multiply',
                'value': 10,
            },
            outputs=[
                dict(
                    uuid='integer_multiply2',
                    column_type='number',
                ),
            ],
        )
        df_new = add_column(add_column(df, action1), action2)
        df_expected = pd.DataFrame(
            [
                [1, 3, 7, 9, 3, 70],
                [4, 2, 9, 3, 8, 90],
            ],
            columns=[
                'integer1',
                'integer2',
                'integer3',
                'integer4',
                'integer_multiply',
                'integer_multiply2',
            ],
        )
        assert_frame_equal(df_new, df_expected)

    def test_add_column_string_replace(self):
        df = pd.DataFrame(
            [
                ['$1000'],
                ['$321.  '],
                ['$4,321'],
            ],
            columns=[
                'amount',
            ],
        )
        action = dict(
            action_arguments=['amount'],
            action_options={
                'udf': 'string_replace',
                'pattern': '\\$|\\.|\\,|\\s*',
                'replacement': '',
            },
            outputs=[
                dict(
                    uuid='amount_clean',
                    column_type='true_or_false',
                ),
            ],
        )
        df_new = add_column(df, action)
        df_expected = pd.DataFrame(
            [
                ['$1000', '1000'],
                ['$321.  ', '321'],
                ['$4,321', '4321'],
            ],
            columns=[
                'amount',
                'amount_clean',
            ],
        )
        assert_frame_equal(df_new, df_expected)

    def test_add_column_string_split(self):
        df = pd.DataFrame(
            [
                ['Street1, Long Beach, CA, '],
                ['Street2,Vernon, CA, 123'],
                ['Pacific Coast Highway, Los Angeles, CA, 111'],
            ],
            columns=[
                'location',
            ],
        )
        action = dict(
            action_arguments=['location'],
            action_options={
                'udf': 'string_split',
                'separator': ',',
                'part_index': 1,
            },
            outputs=[
                dict(
                    uuid='location_city',
                    column_type='text',
                ),
            ],
        )
        action2 = dict(
            action_arguments=['location'],
            action_options={
                'udf': 'string_split',
                'separator': ',',
                'part_index': 3,
            },
            outputs=[
                dict(
                    uuid='num',
                    column_type='number',
                ),
            ],
        )
        df_new = add_column(add_column(df, action), action2)
        df_expected = pd.DataFrame(
            [
                ['Street1, Long Beach, CA, ', 'Long Beach', 0],
                ['Street2,Vernon, CA, 123', 'Vernon', 123],
                ['Pacific Coast Highway, Los Angeles, CA, 111', 'Los Angeles', 111],
            ],
            columns=[
                'location',
                'location_city',
                'num',
            ],
        )
        assert_frame_equal(df_new, df_expected)

    def test_add_column_substring(self):
        df = pd.DataFrame(
            [
                ['$1000.0'],
                ['$321.9'],
            ],
            columns=[
                'amount',
            ],
        )
        action = dict(
            action_arguments=['amount'],
            action_options={
                'udf': 'substring',
                'start': 1,
                'stop': -2,
            },
            outputs=[
                dict(
                    uuid='amount_int',
                    column_type='text',
                ),
            ],
        )
        df_new = add_column(df, action)
        df_expected = pd.DataFrame(
            [
                ['$1000.0', '1000'],
                ['$321.9', '321'],
            ],
            columns=[
                'amount',
                'amount_int',
            ],
        )
        assert_frame_equal(df_new, df_expected)

    def test_average(self):
        from mage_ai.data_cleaner.transformer_actions.column import average

        action = self.__groupby_agg_action('average_amount')
        df_new = average(TEST_DATAFRAME.copy(), action)
        df_expected = pd.DataFrame(
            [
                [1, 1000, 1050],
                [2, 1050, 1100],
                [1, 1100, 1050],
                [2, 1150, 1100],
            ],
            columns=['group_id', 'amount', 'average_amount'],
        )
        df_new['average_amount'] = df_new['average_amount'].astype(int)
        assert_frame_equal(df_new, df_expected)

    def test_count(self):
        df = pd.DataFrame(
            [
                [1, 1000],
                [1, 1050],
                [1, 1100],
                [2, 1150],
            ],
            columns=[
                'group_id',
                'order_id',
            ],
        )
        action = dict(
            action_arguments=['order_id'],
            action_options=dict(groupby_columns=['group_id']),
            outputs=[
                dict(uuid='order_count'),
            ],
        )
        df_new = count(df, action)
        self.assertEqual(
            df_new.to_dict(orient='records'),
            [
                dict(
                    group_id=1,
                    order_id=1000,
                    order_count=3,
                ),
                dict(
                    group_id=1,
                    order_id=1050,
                    order_count=3,
                ),
                dict(
                    group_id=1,
                    order_id=1100,
                    order_count=3,
                ),
                dict(
                    group_id=2,
                    order_id=1150,
                    order_count=1,
                ),
            ],
        )

    def test_count_distinct(self):
        df = pd.DataFrame(
            [
                [1, 1000],
                [1, 1000],
                [1, 1100],
                [2, 1150],
            ],
            columns=[
                'group_id',
                'order_id',
            ],
        )
        action = dict(
            action_arguments=['order_id'],
            action_options=dict(groupby_columns=['group_id']),
            outputs=[
                dict(uuid='order_count'),
            ],
        )
        df_new = count_distinct(df, action)
        self.assertEqual(
            df_new.to_dict(orient='records'),
            [
                dict(
                    group_id=1,
                    order_id=1000,
                    order_count=2,
                ),
                dict(
                    group_id=1,
                    order_id=1000,
                    order_count=2,
                ),
                dict(
                    group_id=1,
                    order_id=1100,
                    order_count=2,
                ),
                dict(
                    group_id=2,
                    order_id=1150,
                    order_count=1,
                ),
            ],
        )

    def test_count_with_time_window(self):
        df = pd.DataFrame(
            [
                [1, 1000, '2021-10-01', '2021-09-01'],
                [1, 1050, '2021-10-01', '2021-08-01'],
                [1, 1100, '2021-10-01', '2021-01-01'],
                [2, 1150, '2021-09-01', '2021-08-01'],
            ],
            columns=[
                'group_id',
                'order_id',
                'group_churned_at',
                'order_created_at',
            ],
        )
        action = dict(
            action_arguments=['order_id'],
            action_code='',
            action_options=dict(
                groupby_columns=['group_id'],
                timestamp_feature_a='group_churned_at',
                timestamp_feature_b='order_created_at',
                window=90 * 24 * 3600,
            ),
            outputs=[
                dict(uuid='order_count'),
            ],
        )
        df_new = count(df, action)
        self.assertEqual(
            df_new.to_dict(orient='records'),
            [
                dict(
                    group_id=1,
                    order_id=1000,
                    group_churned_at='2021-10-01',
                    order_created_at='2021-09-01',
                    order_count=2,
                ),
                dict(
                    group_id=1,
                    order_id=1050,
                    group_churned_at='2021-10-01',
                    order_created_at='2021-08-01',
                    order_count=2,
                ),
                dict(
                    group_id=1,
                    order_id=1100,
                    group_churned_at='2021-10-01',
                    order_created_at='2021-01-01',
                    order_count=2,
                ),
                dict(
                    group_id=2,
                    order_id=1150,
                    group_churned_at='2021-09-01',
                    order_created_at='2021-08-01',
                    order_count=1,
                ),
            ],
        )

    def test_count_with_filter(self):
        df = pd.DataFrame(
            [
                [1, 1000, '2021-10-01', '2021-09-01'],
                [1, 1050, '2021-10-01', '2021-08-01'],
                [1, 1100, '2021-10-01', '2021-01-01'],
                [2, 1150, '2021-09-01', '2021-08-01'],
                [2, 1200, '2021-09-01', '2021-08-16'],
                [2, 1250, '2021-09-01', '2021-08-14'],
            ],
            columns=[
                'group_id',
                'order_id',
                'group_churned_at',
                'order_created_at',
            ],
        )
        action = dict(
            action_arguments=['order_id'],
            action_code='order_created_at < \'2021-08-15\'',
            action_options=dict(
                groupby_columns=['group_id'],
            ),
            outputs=[
                dict(uuid='order_count'),
            ],
        )
        df_new = count(df, action)
        df_expected = pd.DataFrame(
            [
                [1, 1000, '2021-10-01', '2021-09-01', 2],
                [1, 1050, '2021-10-01', '2021-08-01', 2],
                [1, 1100, '2021-10-01', '2021-01-01', 2],
                [2, 1150, '2021-09-01', '2021-08-01', 2],
                [2, 1200, '2021-09-01', '2021-08-16', 2],
                [2, 1250, '2021-09-01', '2021-08-14', 2],
            ],
            columns=[
                'group_id',
                'order_id',
                'group_churned_at',
                'order_created_at',
                'order_count',
            ],
        )
        assert_frame_equal(df_new, df_expected)

    def test_clean_column_name(self):
        df = pd.DataFrame(
            [['', '', '', '', '', '', '', '', '', '', '']],
            columns=[
                'good_name',
                '  Bad Case   ',
                '%@#342%34@@#342',
                ' yield  ',
                '12342   ',
                '1234.  23',
                '   true_crime',
                '@#f$%&*o$*(%^&r*$%&',
                'PascalCaseColumn',
                'camelCaseText',
                '___weird_snake_case___',
            ],
        )
        expected_df = pd.DataFrame(
            [['', '', '', '', '', '', '', '', '', '', '']],
            columns=[
                'good_name',
                'bad_case',
                'number_34234342',
                'yield_',
                'number_12342',
                '1234___23',
                'true_crime',
                'for_',
                'pascal_case_column',
                'camel_case_text',
                'weird_snake_case',
            ],
        )
        action = dict(
            action_type='clean_column_name',
            action_arguments=[
                'good_name',
                '  Bad Case   ',
                '%@#342%34@@#342',
                ' yield  ',
                '12342   ',
                '1234.  23',
                '   true_crime',
                '@#f$%&*o$*(%^&r*$%&',
                'PascalCaseColumn',
                'camelCaseText',
                '___weird_snake_case___',
            ],
            action_code='',
            action_options={},
            action_variables={},
            axis='column',
            outputs=[],
        )
        new_df = clean_column_names(df, action)
        assert_frame_equal(new_df, expected_df)

    def test_diff(self):
        df = pd.DataFrame(
            [
                ['2020-01-01', 1000],
                ['2020-01-02', 1050],
                ['2020-01-03', 1200],
                ['2020-01-04', 990],
            ],
            columns=[
                'date',
                'sold',
            ],
        )
        action = dict(
            action_arguments=['sold'],
            outputs=[
                dict(uuid='sold_diff'),
            ],
        )
        df_new = diff(df, action)
        self.assertEqual(
            df_new.to_dict(orient='records')[1:],
            [
                dict(
                    date='2020-01-02',
                    sold=1050,
                    sold_diff=50,
                ),
                dict(
                    date='2020-01-03',
                    sold=1200,
                    sold_diff=150,
                ),
                dict(
                    date='2020-01-04',
                    sold=990,
                    sold_diff=-210,
                ),
            ],
        )

    # def test_expand_column(self):
    #     df = pd.DataFrame([
    #         [1, 'game'],
    #         [1, 'book'],
    #         [1, 'game'],
    #         [2, 'Video Game'],
    #         [1, 'Video Game'],
    #         [2, 'book'],
    #         [1, 'Video Game'],
    #         [2, 'Video Game'],
    #     ], columns=[
    #         'group_id',
    #         'category',
    #     ])
    #     action = dict(
    #         action_arguments=['category'],
    #         action_options=dict(
    #             groupby_columns=['group_id']
    #         ),
    #         outputs=[
    #             dict(uuid='category_expanded_count_game'),
    #             dict(uuid='category_expanded_count_book'),
    #             dict(uuid='category_expanded_count_video_game'),
    #             dict(uuid='category_expanded_count_clothing'),
    #         ],
    #     )
    #     df_new = expand_column(df, action)
    #     df_expected = pd.DataFrame([
    #         [1, 'game', 2, 1, 2],
    #         [1, 'book', 2, 1, 2],
    #         [1, 'game', 2, 1, 2],
    #         [2, 'Video Game', 0, 1, 2],
    #         [1, 'Video Game', 2, 1, 2],
    #         [2, 'book', 0, 1, 2],
    #         [1, 'Video Game', 2, 1, 2],
    #         [2, 'Video Game', 0, 1, 2],
    #     ], columns=[
    #         'group_id',
    #         'category',
    #         'category_expanded_count_game',
    #         'category_expanded_count_book',
    #         'category_expanded_count_video_game',
    #     ])
    #     assert_frame_equal(df_new, df_expected)

    # def test_expand_column_with_time_window(self):
    #     df = pd.DataFrame([
    #         [1, 'game', '2021-01-02', '2021-01-04'],
    #         [1, 'book', '2021-01-02', '2021-01-04'],
    #         [1, 'game', '2021-01-03', '2021-01-04'],
    #         [2, 'Video Game', '2021-01-01', '2021-01-03'],
    #         [1, 'Video Game', '2021-01-01', '2021-01-04'],
    #         [2, 'book', '2021-01-02', '2021-01-03'],
    #         [1, 'Video Game', '2021-01-03', '2021-01-04'],
    #         [2, 'Video Game', '2020-12-30', '2021-01-03'],
    #     ], columns=[
    #         'group_id',
    #         'category',
    #         'timestamp1',
    #         'timestamp2',
    #     ])
    #     action = dict(
    #         action_arguments=['category'],
    #         action_options=dict(
    #             groupby_columns=['group_id'],
    #             timestamp_feature_a='timestamp2',
    #             timestamp_feature_b='timestamp1',
    #             window=172800,
    #         ),
    #         outputs=[
    #             dict(uuid='category_expanded_count_game_2d'),
    #             dict(uuid='category_expanded_count_book_2d'),
    #             dict(uuid='category_expanded_count_video_game_2d'),
    #             dict(uuid='category_expanded_count_clothing_2d'),
    #         ],
    #     )
    #     df_new = expand_column(df, action)
    #     df_expected = pd.DataFrame([
    #         [1, 'game', '2021-01-02', '2021-01-04', 2, 1, 1],
    #         [1, 'book', '2021-01-02', '2021-01-04', 2, 1, 1],
    #         [1, 'game', '2021-01-03', '2021-01-04', 2, 1, 1],
    #         [2, 'Video Game', '2021-01-01', '2021-01-03', 0, 1, 1],
    #         [1, 'Video Game', '2021-01-01', '2021-01-04', 2, 1, 1],
    #         [2, 'book', '2021-01-02', '2021-01-03', 0, 1, 1],
    #         [1, 'Video Game', '2021-01-03', '2021-01-04', 2, 1, 1],
    #         [2, 'Video Game', '2020-12-30', '2021-01-03', 0, 1, 1],
    #     ], columns=[
    #         'group_id',
    #         'category',
    #         'timestamp1',
    #         'timestamp2',
    #         'category_expanded_count_game_2d',
    #         'category_expanded_count_book_2d',
    #         'category_expanded_count_video_game_2d',
    #     ])
    #     assert_frame_equal(df_new, df_expected)

    def test_first_column(self):
        df = pd.DataFrame(
            [
                [1, 1000],
                [2, 1050],
                [1, 1100],
                [2, 1150],
            ],
            columns=[
                'group_id',
                'order_id',
            ],
        )
        action = dict(
            action_arguments=['order_id'],
            action_options=dict(groupby_columns=['group_id']),
            outputs=[
                dict(uuid='first_order'),
            ],
        )
        df_new = first(df, action)
        self.assertEqual(
            df_new.to_dict(orient='records'),
            [
                dict(
                    group_id=1,
                    order_id=1000,
                    first_order=1000,
                ),
                dict(
                    group_id=2,
                    order_id=1050,
                    first_order=1050,
                ),
                dict(
                    group_id=1,
                    order_id=1100,
                    first_order=1000,
                ),
                dict(
                    group_id=2,
                    order_id=1150,
                    first_order=1050,
                ),
            ],
        )

    def test_fix_syntax_errors(self):
        df = pd.DataFrame(
            [
                [
                    '-12.3%',
                    1000,
                    '11111',
                    'email@maile.com',
                    '12223334444',
                    '12/13/2014',
                    dt(2022, 6, 27),
                ],
                [
                    '$2.234',
                    1050,
                    '2222-2222',
                    'er34ee@int.co',
                    ' 1(000)-111-2222',
                    'not a time',
                    dt(2022, 6, 27),
                ],
                [
                    'eieio',
                    np.nan,
                    '09876',
                    None,
                    '12345678901234',
                    '4/27/2019 12:34:45',
                    dt(2022, 6, 27),
                ],
                ['-4', 1150, None, 'email@email@email.com', 'qqqqqqqqqqq', None, dt(2022, 6, 27)],
                [
                    'not a number',
                    1150,
                    '23423932423',
                    'eeeeeeeee',
                    '1234',
                    '12-24-2022',
                    dt(2022, 6, 27),
                ],
                [
                    None,
                    1150,
                    '234.3324',
                    'agoodemail@network.net',
                    '43213240089',
                    'is not time',
                    dt(2022, 6, 27),
                ],
            ],
            columns=[
                'number',
                'number_but_correct_type',
                'zipcode',
                'email',
                'phone_number',
                'date',
                'date_but_correct_type',
            ],
        )
        action = dict(
            action_arguments=[
                'number',
                'number_but_correct_type',
                'zipcode',
                'email',
                'phone_number',
                'date',
                'date_but_correct_type',
            ],
            action_variables=dict(
                number=dict(feature=dict(column_type='number', uuid='number'), type='feature'),
                number_but_correct_type=dict(
                    feature=dict(column_type='number', uuid='number_but_correct_type'),
                    type='feature',
                ),
                zipcode=dict(feature=dict(column_type='zip_code', uuid='zipcode'), type='feature'),
                email=dict(feature=dict(column_type='email', uuid='email'), type='feature'),
                phone_number=dict(
                    feature=dict(column_type='phone_number', uuid='phone_number'),
                    type='feature',
                ),
                date=dict(feature=dict(column_type='datetime', uuid='date'), type='feature'),
                date_but_correct_type=dict(
                    feature=dict(column_type='datetime', uuid='date_but_correct_type'),
                    type='feature',
                ),
            ),
        )
        new_df = fix_syntax_errors(df, action)
        expected_df = pd.DataFrame(
            [
                [
                    '-12.3%',
                    1000,
                    '11111',
                    'email@maile.com',
                    '12223334444',
                    '12/13/2014',
                    dt(2022, 6, 27),
                ],
                [
                    '$2.234',
                    1050,
                    '2222-2222',
                    'er34ee@int.co',
                    ' 1(000)-111-2222',
                    pd.NaT,
                    dt(2022, 6, 27),
                ],
                [
                    np.nan,
                    np.nan,
                    '09876',
                    None,
                    'invalid',
                    '4/27/2019 12:34:45',
                    dt(2022, 6, 27),
                ],
                ['-4', 1150, None, 'invalid', 'invalid', None, dt(2022, 6, 27)],
                [
                    np.nan,
                    1150,
                    'invalid',
                    'invalid',
                    'invalid',
                    '12-24-2022',
                    dt(2022, 6, 27),
                ],
                [
                    None,
                    1150,
                    'invalid',
                    'agoodemail@network.net',
                    '43213240089',
                    pd.NaT,
                    dt(2022, 6, 27),
                ],
            ],
            columns=[
                'number',
                'number_but_correct_type',
                'zipcode',
                'email',
                'phone_number',
                'date',
                'date_but_correct_type',
            ],
        )
        assert_frame_equal(new_df, expected_df)

    def test_impute(self):
        from mage_ai.data_cleaner.transformer_actions.column import impute

        df = pd.DataFrame(
            [
                ['2020-01-01', 1000, '       ', 800],
                ['2020-01-03', '', 1200, 700],
                ['2020-01-05', 1200, np.NaN, 900],
                ['2020-01-02', np.NaN, '  ', 700],
                ['2020-01-04', 1700, 1300, 800],
            ],
            columns=[
                'date',
                'sold',
                'curr_profit',
                'prev_sold',
            ],
        )
        action1 = dict(
            action_arguments=['sold', 'curr_profit'],
            action_options={
                'value': '0',
            },
            action_variables={
                'sold': {
                    'feature': {
                        'column_type': 'number',
                        'uuid': 'sold',
                    },
                    'type': 'feature',
                },
                'curr_profit': {
                    'feature': {
                        'column_type': 'number',
                        'uuid': 'curr_profit',
                    },
                    'type': 'feature',
                },
            },
        )
        action2 = dict(
            action_arguments=['sold'],
            action_options={
                'value': '0',
            },
            action_variables={
                'sold': {
                    'feature': {
                        'column_type': 'number',
                        'uuid': 'sold',
                    },
                    'type': 'feature',
                },
                'curr_profit': {
                    'feature': {
                        'column_type': 'number',
                        'uuid': 'curr_profit',
                    },
                    'type': 'feature',
                },
            },
        )
        action3 = dict(
            action_arguments=['sold', 'curr_profit'],
            action_options={
                'strategy': 'average',
            },
            action_variables={
                'sold': {
                    'feature': {
                        'column_type': 'number',
                        'uuid': 'sold',
                    },
                    'type': 'feature',
                },
                'curr_profit': {
                    'feature': {
                        'column_type': 'number',
                        'uuid': 'curr_profit',
                    },
                    'type': 'feature',
                },
            },
        )
        action4 = dict(
            action_arguments=['sold', 'curr_profit'],
            action_options={
                'strategy': 'median',
            },
            action_variables={
                'sold': {
                    'feature': {
                        'column_type': 'number',
                        'uuid': 'sold',
                    },
                    'type': 'feature',
                },
                'curr_profit': {
                    'feature': {
                        'column_type': 'number',
                        'uuid': 'curr_profit',
                    },
                    'type': 'feature',
                },
            },
        )
        action5 = dict(
            action_arguments=['sold', 'curr_profit'],
            action_options={
                'strategy': 'column',
                'value': 'prev_sold',
            },
            action_variables={
                'sold': {
                    'feature': {
                        'column_type': 'number',
                        'uuid': 'sold',
                    },
                    'type': 'feature',
                },
                'curr_profit': {
                    'feature': {
                        'column_type': 'number',
                        'uuid': 'curr_profit',
                    },
                    'type': 'feature',
                },
            },
        )
        action6 = dict(
            action_arguments=['sold', 'curr_profit'],
            action_options={'strategy': 'sequential', 'timeseries_index': ['date']},
            action_variables={
                'sold': {
                    'feature': {
                        'column_type': 'number',
                        'uuid': 'sold',
                    },
                    'type': 'feature',
                },
                'curr_profit': {
                    'feature': {
                        'column_type': 'number',
                        'uuid': 'curr_profit',
                    },
                    'type': 'feature',
                },
            },
        )
        action7 = dict(
            action_arguments=['sold', 'curr_profit'],
            action_options={'strategy': 'random'},
            action_variables={
                'sold': {
                    'feature': {
                        'column_type': 'number',
                        'uuid': 'sold',
                    },
                    'type': 'feature',
                },
                'curr_profit': {
                    'feature': {
                        'column_type': 'number',
                        'uuid': 'curr_profit',
                    },
                    'type': 'feature',
                },
            },
        )
        action8 = dict(
            action_arguments=['sold', 'curr_profit'],
            action_options={'strategy': 'mode'},
            action_variables={
                'sold': {
                    'feature': {
                        'column_type': 'number',
                        'uuid': 'sold',
                    },
                    'type': 'feature',
                },
                'curr_profit': {
                    'feature': {
                        'column_type': 'number',
                        'uuid': 'curr_profit',
                    },
                    'type': 'feature',
                },
            },
        )
        action_invalid = dict(
            action_arguments=['sold', 'curr_profit'],
            action_options={
                'strategy': 'knn',
            },
            action_variables={
                'sold': {
                    'feature': {
                        'column_type': 'number',
                        'uuid': 'sold',
                    },
                    'type': 'feature',
                },
                'curr_profit': {
                    'feature': {
                        'column_type': 'number',
                        'uuid': 'curr_profit',
                    },
                    'type': 'feature',
                },
            },
        )
        df_new1 = impute(df.copy(), action1)
        df_new2 = impute(df.copy(), action2)
        df_new3 = impute(df.copy(), action3)
        df_new4 = impute(df.copy(), action4)
        df_new5 = impute(df.copy(), action5)
        df_new6 = impute(df.copy(), action6).reset_index(drop=True)
        df_new7 = impute(df.copy(), action7)
        df_new8 = impute(df.copy(), action8)

        df_expected1 = pd.DataFrame(
            [
                ['2020-01-01', 1000, 0, 800],
                ['2020-01-03', 0, 1200, 700],
                ['2020-01-05', 1200, 0, 900],
                ['2020-01-02', 0, 0, 700],
                ['2020-01-04', 1700, 1300, 800],
            ],
            columns=[
                'date',
                'sold',
                'curr_profit',
                'prev_sold',
            ],
        )
        df_expected2 = pd.DataFrame(
            [
                ['2020-01-01', 1000, '       ', 800],
                ['2020-01-03', 0, 1200, 700],
                ['2020-01-05', 1200, np.nan, 900],
                ['2020-01-02', 0, '  ', 700],
                ['2020-01-04', 1700, 1300, 800],
            ],
            columns=[
                'date',
                'sold',
                'curr_profit',
                'prev_sold',
            ],
        )
        df_expected3 = pd.DataFrame(
            [
                ['2020-01-01', 1000, 1250, 800],
                ['2020-01-03', 1300, 1200, 700],
                ['2020-01-05', 1200, 1250, 900],
                ['2020-01-02', 1300, 1250, 700],
                ['2020-01-04', 1700, 1300, 800],
            ],
            columns=[
                'date',
                'sold',
                'curr_profit',
                'prev_sold',
            ],
        )
        df_expected4 = pd.DataFrame(
            [
                ['2020-01-01', 1000, 1250, 800],
                ['2020-01-03', 1200, 1200, 700],
                ['2020-01-05', 1200, 1250, 900],
                ['2020-01-02', 1200, 1250, 700],
                ['2020-01-04', 1700, 1300, 800],
            ],
            columns=[
                'date',
                'sold',
                'curr_profit',
                'prev_sold',
            ],
        )
        df_expected5 = pd.DataFrame(
            [
                ['2020-01-01', 1000, 800, 800],
                ['2020-01-03', 700, 1200, 700],
                ['2020-01-05', 1200, 900, 900],
                ['2020-01-02', 700, 700, 700],
                ['2020-01-04', 1700, 1300, 800],
            ],
            columns=[
                'date',
                'sold',
                'curr_profit',
                'prev_sold',
            ],
        )
        df_expected6 = pd.DataFrame(
            [
                ['2020-01-01', 1000, 0, 800],
                ['2020-01-02', 1000, 0, 700],
                ['2020-01-03', 1000, 1200, 700],
                ['2020-01-04', 1700, 1300, 800],
                ['2020-01-05', 1200, 1300, 900],
            ],
            columns=[
                'date',
                'sold',
                'curr_profit',
                'prev_sold',
            ],
        )
        df_expected8 = pd.DataFrame(
            [
                ['2020-01-01', 1000, 1200, 800],
                ['2020-01-03', 1000, 1200, 700],
                ['2020-01-05', 1200, 1200, 900],
                ['2020-01-02', 1000, 1200, 700],
                ['2020-01-04', 1700, 1300, 800],
            ],
            columns=[
                'date',
                'sold',
                'curr_profit',
                'prev_sold',
            ],
        )

        df_new1['sold'] = df_new1['sold'].astype(int)
        df_new1['curr_profit'] = df_new1['curr_profit'].astype(int)
        df_new2['sold'] = df_new2['sold'].astype(int)
        df_new3['sold'] = df_new3['sold'].astype(int)
        df_new3['curr_profit'] = df_new3['curr_profit'].astype(int)
        df_new4['sold'] = df_new4['sold'].astype(int)
        df_new4['curr_profit'] = df_new4['curr_profit'].astype(int)
        df_new5['sold'] = df_new5['sold'].astype(int)
        df_new5['curr_profit'] = df_new5['curr_profit'].astype(int)
        df_new6['sold'] = df_new6['sold'].astype(int)
        df_new7['sold'] = df_new7['sold'].astype(int)
        df_new7['curr_profit'] = df_new7['curr_profit'].astype(int)
        df_new8['sold'] = df_new8['sold'].astype(int)
        df_new8['curr_profit'] = df_new8['curr_profit'].astype(int)

        assert_frame_equal(df_new1, df_expected1)
        assert_frame_equal(df_new2, df_expected2)
        assert_frame_equal(df_new3, df_expected3)
        assert_frame_equal(df_new4, df_expected4)
        assert_frame_equal(df_new5, df_expected5)
        assert_frame_equal(df_new6, df_expected6)
        assert_frame_equal(df_new7, df_new7.dropna(axis=0))
        assert_frame_equal(df_new8, df_expected8)

        with self.assertRaises(Exception):
            _ = impute(df.copy(), action_invalid)

    def test_impute_random_edge(self):
        from mage_ai.data_cleaner.transformer_actions.column import impute

        df = pd.DataFrame(
            [
                ['2020-01-01', 1000, '       ', 800],
                ['2020-01-02', '', None, 700],
                ['2020-01-03', 1200, np.NaN, 900],
                ['2020-01-04', np.NaN, '  ', 700],
                ['2020-01-05', 1700, np.NaN, 800],
            ],
            columns=[
                'date',
                'sold',
                'curr_profit',
                'prev_sold',
            ],
        )
        action = dict(
            action_arguments=['sold', 'curr_profit'],
            action_options={'strategy': 'random'},
        )
        with self.assertRaises(Exception):
            _ = impute(df.copy(), action)

    def test_impute_constant(self):
        from mage_ai.data_cleaner.transformer_actions.column import impute

        df = pd.DataFrame(
            [
                [1, 1.000, '2021-10-01', 'Store 1', 23023],
                [1, None, '2021-10-01', 'Store 2', np.nan],
                [np.nan, 1100, '', '', 90233],
                [2, None, None, 'Store 1', 23920],
                [2, 12.00, '2021-09-01', None, np.nan],
                [2, 125.0, '2021-09-01', 'Store 3', 49833],
            ],
            columns=[
                'group_id',
                'price',
                'group_churned_at',
                'store',
                'zip_code',
            ],
        )
        action = dict(
            action_arguments=['group_id', 'price', 'group_churned_at', 'store', 'zip_code'],
            action_options={
                'strategy': 'constant',
            },
            action_variables={
                'group_id': {
                    'feature': {
                        'column_type': 'number',
                        'uuid': 'group_id',
                    },
                    'type': 'feature',
                },
                'price': {
                    'feature': {
                        'column_type': 'number_with_decimals',
                        'uuid': 'price',
                    },
                    'type': 'feature',
                },
                'group_churned_at': {
                    'feature': {
                        'column_type': 'datetime',
                        'uuid': 'group_churned_at',
                    },
                    'type': 'feature',
                },
                'store': {
                    'feature': {
                        'column_type': 'category',
                        'uuid': 'store',
                    },
                    'type': 'feature',
                },
                'zip_code': {
                    'feature': {
                        'column_type': 'zip_code',
                        'uuid': 'zip_code',
                    },
                    'type': 'feature',
                },
            },
        )
        df_expected = pd.DataFrame(
            [
                [1, 1.000, '2021-10-01', 'Store 1', 23023],
                [1, 0.0, '2021-10-01', 'Store 2', 'missing'],
                [0, 1100, pd.Timestamp.min, 'missing', 90233],
                [2, 0.0, pd.Timestamp.min, 'Store 1', 23920],
                [2, 12.00, '2021-09-01', 'missing', 'missing'],
                [2, 125.0, '2021-09-01', 'Store 3', 49833],
            ],
            columns=[
                'group_id',
                'price',
                'group_churned_at',
                'store',
                'zip_code',
            ],
        )
        df_new = impute(df, action).reset_index(drop=True)
        df_new['group_id'] = df_new['group_id'].astype(int)
        df_new['price'] = df_new['price'].astype(float)
        df_new['group_churned_at'] = df_new['group_churned_at'].astype(np.datetime64)
        assert_frame_equal(df_new, df_expected)

    def test_impute_constant_with_value(self):
        from mage_ai.data_cleaner.transformer_actions.column import impute

        df = pd.DataFrame(
            [
                [1, 1.000, '2021-10-01', 'Store 1', 23023],
                [1, None, '2021-10-01', 'Store 2', np.nan],
                [np.nan, 1100, '', '', 90233],
                [2, None, None, 'Store 1', 23920],
                [2, 12.00, '2021-09-01', None, np.nan],
                [2, 125.0, '2021-09-01', 'Store 3', 49833],
            ],
            columns=[
                'group_id',
                'price',
                'group_churned_at',
                'store',
                'zip_code',
            ],
        )
        action = dict(
            action_arguments=['group_id', 'price', 'group_churned_at', 'store', 'zip_code'],
            action_options={'strategy': 'constant', 'value': 0},
            action_variables={
                'group_id': {
                    'feature': {'column_type': 'number', 'uuid': 'group_id'},
                    'type': 'feature',
                },
                'price': {
                    'feature': {'column_type': 'number_with_decimals', 'uuid': 'price'},
                    'type': 'feature',
                },
                'group_churned_at': {
                    'feature': {'column_type': 'datetime', 'uuid': 'group_churned_at'},
                    'type': 'feature',
                },
                'store': {
                    'feature': {'column_type': 'category', 'uuid': 'store'},
                    'type': 'feature',
                },
                'zip_code': {
                    'feature': {'column_type': 'zip_code', 'uuid': 'zip_code'},
                    'type': 'feature',
                },
            },
        )
        df_expected = pd.DataFrame(
            [
                [1, 1.000, '2021-10-01', 'Store 1', 23023],
                [1, 0, '2021-10-01', 'Store 2', 0],
                [0, 1100, 0, 0, 90233],
                [2, 0, 0, 'Store 1', 23920],
                [2, 12.00, '2021-09-01', 0, 0],
                [2, 125.0, '2021-09-01', 'Store 3', 49833],
            ],
            columns=[
                'group_id',
                'price',
                'group_churned_at',
                'store',
                'zip_code',
            ],
        )
        df_new = impute(df, action).reset_index(drop=True)
        df_new['zip_code'] = df_new['zip_code'].astype(int)
        assert_frame_equal(df_new, df_expected)

    def test_impute_sequential_two_idx(self):
        from mage_ai.data_cleaner.transformer_actions.column import impute

        df = pd.DataFrame(
            [
                [1, 1000, '2021-10-01', '2021-09-01', 2],
                [1, None, '2021-10-01', '2021-08-01', np.nan],
                [np.nan, 1100, '2021-10-01', '2021-01-01', 2],
                [2, None, '2021-09-01', '2021-08-01', 2],
                [2, 1200, '2021-09-01', '2021-08-16', np.nan],
                [2, 1250, '2021-09-01', '2021-08-14', 2],
            ],
            columns=[
                'group_id',
                'order_id',
                'group_churned_at',
                'order_created_at',
                'order_count',
            ],
        )
        action = dict(
            action_arguments=['group_id', 'order_id', 'order_count'],
            action_options={
                'strategy': 'sequential',
                'timeseries_index': ['group_churned_at', 'order_created_at'],
            },
            action_variables={
                'group_id': {
                    'feature': {
                        'column_type': 'number',
                        'uuid': 'group_id',
                    },
                    'type': 'feature',
                },
                'order_id': {
                    'feature': {
                        'column_type': 'number',
                        'uuid': 'order_id',
                    },
                    'type': 'feature',
                },
                'group_churned_at': {
                    'feature': {
                        'column_type': 'datetime',
                        'uuid': 'group_churned_at',
                    },
                    'type': 'feature',
                },
                'order_created_at': {
                    'feature': {
                        'column_type': 'datetime',
                        'uuid': 'order_created_at',
                    },
                    'type': 'feature',
                },
                'order_count': {
                    'feature': {
                        'column_type': 'number',
                        'uuid': 'order_count',
                    },
                    'type': 'feature',
                },
            },
        )
        df_expected = pd.DataFrame(
            [
                [2, 0, '2021-09-01', '2021-08-01', 2],
                [2, 1250, '2021-09-01', '2021-08-14', 2],
                [2, 1200, '2021-09-01', '2021-08-16', 2],
                [2, 1100, '2021-10-01', '2021-01-01', 2],
                [1, 1100, '2021-10-01', '2021-08-01', 2],
                [1, 1000, '2021-10-01', '2021-09-01', 2],
            ],
            columns=[
                'group_id',
                'order_id',
                'group_churned_at',
                'order_created_at',
                'order_count',
            ],
        )
        df_new = impute(df, action).reset_index(drop=True)
        df_new['group_id'] = df_new['group_id'].astype(int)
        df_new['order_count'] = df_new['order_count'].astype(int)
        assert_frame_equal(df_new, df_expected)

    def test_impute_lists(self):
        from mage_ai.data_cleaner.transformer_actions.column import impute

        df = pd.DataFrame(
            {
                'lists': [
                    ['this', 'is', 'a', 'list', 'of', 'strings'],
                    ['this', 'is', 'a', 'list', 'of', 'strings'],
                    ['this', 'is', 'a', 'list', 'of', 'strings'],
                    None,
                ],
                'lists2': [
                    [2, 1, 3, 4, 2, 1, 2, 2],
                    [8, 9, 6, 4, 6, 4, 5, 4, 3, 4],
                    [],
                    [2, 3, 4, 1, 2],
                ],
                'lists3': [
                    None,
                    [True, False, True, True],
                    [False, True, False, True],
                    [True, True, True, False],
                ],
                'lists4': [
                    [2, 'string', False, None],
                    [np.nan, 2.0, 'string', '3'],
                    ['not string?', True, True, 8, False, np.nan, None],
                    [],
                ],
                'tuples': [
                    (2, 'string', False, None),
                    (np.nan, 2.0, 'string', '3'),
                    ('not string?', True, True, 8, False, np.nan, None),
                    tuple(),
                ],
                'string_lists': [
                    '[2, \'string\', False, None]',
                    None,
                    '[\'not string?\'   ,  True, True , 8   , False  , np.nan, None]',
                    '[]',
                ],
                'string_tuples': [
                    '(2, \'string\', False, None)',
                    '(np.nan, 2.0, \'string\', \'3\')',
                    '(\'not string?\', True, True, 8, False, np.nan, None)',
                    None,
                ],
                'not_a_list': [
                    '3',
                    '4',
                    None,
                    'a very long piece of text',
                ],
            }
        )

        # action_mode = dict(
        #     action_arguments=[
        #         'lists',
        #         'lists2',
        #         'lists3',
        #         'lists4',
        #         'tuples',
        #         'string_lists',
        #         'string_tuples',
        #         'not_a_list',
        #     ],
        #     action_options={'strategy': 'mode'},
        #     action_variables=dict(
        #         lists=dict(feature=dict(column_type='list', uuid='lists'), type='feature'),
        #         lists2=dict(feature=dict(column_type='list', uuid='lists2'), type='feature'),
        #         lists3=dict(feature=dict(column_type='list', uuid='lists3'), type='feature'),
        #         lists4=dict(feature=dict(column_type='list', uuid='lists4'), type='feature'),
        #         tuples=dict(feature=dict(column_type='list', uuid='tuples'), type='feature'),
        #         string_lists=dict(
        #             feature=dict(column_type='list', uuid='string_lists'), type='feature'
        #         ),
        #         string_tuples=dict(
        #             feature=dict(column_type='list', uuid='string_tuples'), type='feature'
        #         ),
        #         not_a_list=dict(
        #             feature=dict(column_type='text', uuid='not_a_list'), type='feature'
        #         ),
        #     ),
        # )
        action_constant = dict(
            action_arguments=[
                'lists',
                'lists2',
                'lists3',
                'lists4',
                'tuples',
                'string_lists',
                'string_tuples',
                'not_a_list',
            ],
            action_options={'strategy': 'constant'},
            action_variables=dict(
                lists=dict(feature=dict(column_type='list', uuid='lists'), type='feature'),
                lists2=dict(feature=dict(column_type='list', uuid='lists2'), type='feature'),
                lists3=dict(feature=dict(column_type='list', uuid='lists3'), type='feature'),
                lists4=dict(feature=dict(column_type='list', uuid='lists4'), type='feature'),
                tuples=dict(feature=dict(column_type='list', uuid='tuples'), type='feature'),
                string_lists=dict(
                    feature=dict(column_type='list', uuid='string_lists'), type='feature'
                ),
                string_tuples=dict(
                    feature=dict(column_type='list', uuid='string_tuples'), type='feature'
                ),
                not_a_list=dict(
                    feature=dict(column_type='text', uuid='not_a_list'), type='feature'
                ),
            ),
        )
        # expected_df_mode = pd.DataFrame(
        #     {
        #         'lists': [
        #             ['this', 'is', 'a', 'list', 'of', 'strings'],
        #             ['this', 'is', 'a', 'list', 'of', 'strings'],
        #             ['this', 'is', 'a', 'list', 'of', 'strings'],
        #             ['this', 'is', 'a', 'list', 'of', 'strings'],
        #         ],
        #         'lists2': [
        #             [2, 1, 3, 4, 2, 1, 2, 2],
        #             [8, 9, 6, 4, 6, 4, 5, 4, 3, 4],
        #             [],
        #             [2, 3, 4, 1, 2],
        #         ],
        #         'lists3': [
        #             [False, True, False, True],
        #             [True, False, True, True],
        #             [False, True, False, True],
        #             [True, True, True, False],
        #         ],
        #         'lists4': [
        #             [2, 'string', False, None],
        #             [np.nan, 2.0, 'string', '3'],
        #             ['not string?', True, True, 8, False, np.nan, None],
        #             [],
        #         ],
        #         'tuples': [
        #             (2, 'string', False, None),
        #             (np.nan, 2.0, 'string', '3'),
        #             ('not string?', True, True, 8, False, np.nan, None),
        #             tuple(),
        #         ],
        #         'string_lists': [
        #             [2, 'string', False, None],
        #             ['not string?', True, True, 8, False, np.nan, None],
        #             ['not string?', True, True, 8, False, np.nan, None],
        #             [],
        #         ],
        #         'string_tuples': [
        #             [2, 'string', False, None],
        #             [np.nan, 2.0, 'string', '3'],
        #             ['not string?', True, True, 8, False, np.nan, None],
        #             ['not string?', True, True, 8, False, np.nan, None],
        #         ],
        #         'not_a_list': [
        #             '3',
        #             '4',
        #             '3',
        #             'a very long piece of text',
        #         ],
        #     }
        # )
        expected_df_constant = pd.DataFrame(
            {
                'lists': [
                    ['this', 'is', 'a', 'list', 'of', 'strings'],
                    ['this', 'is', 'a', 'list', 'of', 'strings'],
                    ['this', 'is', 'a', 'list', 'of', 'strings'],
                    [],
                ],
                'lists2': [
                    [2, 1, 3, 4, 2, 1, 2, 2],
                    [8, 9, 6, 4, 6, 4, 5, 4, 3, 4],
                    [],
                    [2, 3, 4, 1, 2],
                ],
                'lists3': [
                    [],
                    [True, False, True, True],
                    [False, True, False, True],
                    [True, True, True, False],
                ],
                'lists4': [
                    [2, 'string', False, None],
                    [np.nan, 2.0, 'string', '3'],
                    ['not string?', True, True, 8, False, np.nan, None],
                    [],
                ],
                'tuples': [
                    (2, 'string', False, None),
                    (np.nan, 2.0, 'string', '3'),
                    ('not string?', True, True, 8, False, np.nan, None),
                    tuple(),
                ],
                'string_lists': [
                    [2, 'string', False, None],
                    [],
                    ['not string?', True, True, 8, False, np.nan, None],
                    [],
                ],
                'string_tuples': [
                    [2, 'string', False, None],
                    [np.nan, 2.0, 'string', '3'],
                    ['not string?', True, True, 8, False, np.nan, None],
                    [],
                ],
                'not_a_list': [
                    '3',
                    '4',
                    'missing',
                    'a very long piece of text',
                ],
            }
        )
        # new_df_mode = impute(df.copy(), action_mode)
        new_df_constant = impute(df.copy(), action_constant)
        # assert_frame_equal(new_df_mode, expected_df_mode)
        assert_frame_equal(new_df_constant, expected_df_constant)

    def test_impute_list_seq(self):
        from mage_ai.data_cleaner.transformer_actions.column import impute

        df = pd.DataFrame(
            [
                ['CT', '06902', '12-24-2022'],
                ['NY', '10001', '12-25-2022'],
                ['CA', '', '12-28-2022'],
                [None, '', '12-30-2022'],
                ['CA', None, '12-30-2022'],
                ['MA', '12214', '12-31-2022'],
                ['PA', '', '1-2-2023'],
                ['TX', '75001', '1-2-2023'],
                ['', None, ''],
                [None, '', None],
            ],
            columns=['state', 'location', 'timestamp'],
        )
        df['lists'] = pd.Series(
            [
                (np.nan, 2.0, 'string', '3'),
                ['not string?', True, True, 8, False, np.nan, None],
                None,
                None,
                ('not string?', True, True, 8, False, np.nan, None),
                [],
                '[\'not string?\'   ,  True, True , 8   , False  , np.nan, None]',
                '(\'not string?\'   ,  True, True , 8   , False  , np.nan, None)',
                tuple(),
                None,
            ]
        )
        df['timestamp'] = pd.to_datetime(
            df['timestamp'], infer_datetime_format=True, errors='coerce'
        )
        action = dict(
            action_type='impute',
            action_arguments=['state', 'location', 'timestamp', 'lists'],
            action_options=dict(strategy='sequential', timeseries_index=['timestamp']),
            action_variables=dict(
                state=dict(feature=dict(column_type='category', uuid='state'), type='feature'),
                location=dict(
                    feature=dict(column_type='zip_code', uuid='location'), type='feature'
                ),
                timestamp=dict(
                    feature=dict(column_type='datetime', uuid='timestamp'), type='feature'
                ),
                lists=dict(feature=dict(column_type='list', uuid='lists'), type='feature'),
            ),
            action_code='',
            axis='column',
            outputs=[],
        )
        expected_df = pd.DataFrame(
            [
                ['CT', '06902', '12-24-2022'],
                ['NY', '10001', '12-25-2022'],
                ['CA', '10001', '12-28-2022'],
                ['CA', '10001', '12-30-2022'],
                ['CA', '10001', '12-30-2022'],
                ['MA', '12214', '12-31-2022'],
                ['PA', '12214', '1-2-2023'],
                ['TX', '75001', '1-2-2023'],
                ['TX', '75001', '1-2-2023'],
                ['TX', '75001', '1-2-2023'],
            ],
            columns=['state', 'location', 'timestamp'],
        )
        expected_df['lists'] = pd.Series(
            [
                (np.nan, 2.0, 'string', '3'),
                ['not string?', True, True, 8, False, np.nan, None],
                ['not string?', True, True, 8, False, np.nan, None],
                ['not string?', True, True, 8, False, np.nan, None],
                ['not string?', True, True, 8, False, np.nan, None],
                [],
                ['not string?', True, True, 8, False, np.nan, None],
                ['not string?', True, True, 8, False, np.nan, None],
                [],
                [],
            ]
        )
        new_df = impute(df, action).reset_index(drop=True)
        expected_df['timestamp'] = pd.to_datetime(
            expected_df['timestamp'], infer_datetime_format=True, errors='coerce'
        )
        assert_frame_equal(expected_df, new_df)

    def test_last_column(self):
        df = pd.DataFrame(
            [
                [1, 1000],
                [2, 1050],
                [1, 1100],
                [2, 1150],
            ],
            columns=[
                'group_id',
                'order_id',
            ],
        )
        action = dict(
            action_arguments=['order_id'],
            action_options=dict(groupby_columns=['group_id']),
            outputs=[
                dict(uuid='last_order'),
            ],
        )
        df_new = last(df, action)
        self.assertEqual(
            df_new.to_dict(orient='records'),
            [
                dict(
                    group_id=1,
                    order_id=1000,
                    last_order=1100,
                ),
                dict(
                    group_id=2,
                    order_id=1050,
                    last_order=1150,
                ),
                dict(
                    group_id=1,
                    order_id=1100,
                    last_order=1100,
                ),
                dict(
                    group_id=2,
                    order_id=1150,
                    last_order=1150,
                ),
            ],
        )

    def test_max(self):
        from mage_ai.data_cleaner.transformer_actions.column import max

        action = self.__groupby_agg_action('max_amount')
        df_new = max(TEST_DATAFRAME.copy(), action)
        df_expected = pd.DataFrame(
            [
                [1, 1000, 1100],
                [2, 1050, 1150],
                [1, 1100, 1100],
                [2, 1150, 1150],
            ],
            columns=[
                'group_id',
                'amount',
                'max_amount',
            ],
        )
        assert_frame_equal(df_new, df_expected)

        action2 = dict(
            action_arguments=['amount'],
            action_options=dict(),
            outputs=[
                dict(uuid='max_amount'),
            ],
        )
        df_new2 = max(TEST_DATAFRAME.copy(), action2)
        df_expected2 = pd.DataFrame(
            [
                [1, 1000, 1150],
                [2, 1050, 1150],
                [1, 1100, 1150],
                [2, 1150, 1150],
            ],
            columns=[
                'group_id',
                'amount',
                'max_amount',
            ],
        )
        assert_frame_equal(df_new2, df_expected2)

    def test_median(self):
        from mage_ai.data_cleaner.transformer_actions.column import median

        action = self.__groupby_agg_action('median_amount')
        df = pd.DataFrame(
            [
                [1, 1000],
                [2, 1050],
                [1, 1100],
                [2, 1550],
                [2, 1150],
            ],
            columns=[
                'group_id',
                'amount',
            ],
        )
        df_new = median(df, action)
        df_new['median_amount'] = df_new['median_amount'].astype(int)
        df_expected = pd.DataFrame(
            [
                [1, 1000, 1050],
                [2, 1050, 1150],
                [1, 1100, 1050],
                [2, 1550, 1150],
                [2, 1150, 1150],
            ],
            columns=[
                'group_id',
                'amount',
                'median_amount',
            ],
        )
        assert_frame_equal(df_new, df_expected)

    def test_min(self):
        from mage_ai.data_cleaner.transformer_actions.column import min

        action = self.__groupby_agg_action('min_amount')
        df_new = min(TEST_DATAFRAME.copy(), action)
        df_expected = pd.DataFrame(
            [
                [1, 1000, 1000],
                [2, 1050, 1050],
                [1, 1100, 1000],
                [2, 1150, 1050],
            ],
            columns=[
                'group_id',
                'amount',
                'min_amount',
            ],
        )
        assert_frame_equal(df_new, df_expected)

    def test_reformat_capitalization(self):
        from mage_ai.data_cleaner.transformer_actions.column import reformat

        df = pd.DataFrame(
            [
                [None, 'us', 30000, 'Funny Video Corp', 'cute animal #1', 100, 30],
                ['500', 'CA', 10000, 'Machine Learning 4 U', 'intro to regression', 3000, 20],
                ['', '', np.nan, 'News Inc', 'Daily news #1', None, 75],
                ['250', 'CA', 7500, 'Machine Learning 4 U', 'MACHINE LEARNING SEMINAR', 8000, 20],
                ['1000', 'mx', 45003, None, 'cute Animal #4', 90, 40],
                ['1500', 'MX', 75000, 'Funny Video Corp', '', 70, 25],
                ['1500', np.nan, 75000, 'News Inc', 'daily news #3', 70, 25],
                [None, 'mx', 75000, 'Z Combinator', 'Tutorial: how to Start a startup', 70, np.nan],
                ['1250', 'US', 60000, 'Funny Video Corp', 'cute animal #3', 80, 20],
                ['', 'CA', 5000, '', '', 10000, 30],
                ['800', None, 12050, 'Funny Video Corp', 'meme Compilation', 2000, 45],
                ['600', 'CA', 11000, 'News Inc', 'daily news #2', 3000, 50],
                ['600', 'ca', '', 'Funny Video Corp', '', 3000, None],
                ['700', 'MX', 11750, 'Funny Video Corp', 'cute animal #2', 2750, 55],
                ['700', '', None, 'Funny Video Corp', '', None, 55],
                ['700', 'MX', 11750, 'Funny Video Corp', '', 2750, 55],
                ['1200', 'MX', 52000, 'Z Combinator', 'vc funding strats', 75, 60],
            ],
            columns=[
                'userid',
                'location',
                'number_of_creators',
                'company_name',
                'name',
                'losses',
                'number_of_advertisers',
            ],
        )
        df_expected = pd.DataFrame(
            [
                [None, 'US', 30000, 'funny video corp', 'cute animal #1', 100, 30],
                ['500', 'CA', 10000, 'machine learning 4 u', 'intro to regression', 3000, 20],
                ['', None, np.nan, 'news inc', 'daily news #1', None, 75],
                ['250', 'CA', 7500, 'machine learning 4 u', 'machine learning seminar', 8000, 20],
                ['1000', 'MX', 45003, None, 'cute animal #4', 90, 40],
                ['1500', 'MX', 75000, 'funny video corp', None, 70, 25],
                ['1500', np.nan, 75000, 'news inc', 'daily news #3', 70, 25],
                [None, 'MX', 75000, 'z combinator', 'tutorial: how to start a startup', 70, np.nan],
                ['1250', 'US', 60000, 'funny video corp', 'cute animal #3', 80, 20],
                ['', 'CA', 5000, None, None, 10000, 30],
                ['800', None, 12050, 'funny video corp', 'meme compilation', 2000, 45],
                ['600', 'CA', 11000, 'news inc', 'daily news #2', 3000, 50],
                ['600', 'CA', '', 'funny video corp', None, 3000, None],
                ['700', 'MX', 11750, 'funny video corp', 'cute animal #2', 2750, 55],
                ['700', None, None, 'funny video corp', None, None, 55],
                ['700', 'MX', 11750, 'funny video corp', None, 2750, 55],
                ['1200', 'MX', 52000, 'z combinator', 'vc funding strats', 75, 60],
            ],
            columns=[
                'userid',
                'location',
                'number_of_creators',
                'company_name',
                'name',
                'losses',
                'number_of_advertisers',
            ],
        )
        action1 = dict(
            action_type='reformat',
            action_arguments=['location'],
            axis='column',
            action_options={'reformat': 'caps_standardization', 'capitalization': 'uppercase'},
            action_variables={
                'location': {
                    'feature': {'column_type': 'category', 'uuid': 'location'},
                    'type': 'feature',
                }
            },
            action_code='',
            outputs=[],
        )
        action2 = dict(
            action_type='reformat',
            action_arguments=['company_name', 'name'],
            axis='column',
            action_options={'reformat': 'caps_standardization', 'capitalization': 'lowercase'},
            action_variables={
                'company_name': {
                    'feature': {'column_type': 'category_high_cardinality', 'uuid': 'company_name'},
                    'type': 'feature',
                },
                'name': {'feature': {'column_type': 'text', 'uuid': 'name'}, 'type': 'feature'},
            },
            action_code='',
            outputs=[],
        )
        df_new = reformat(df, action1)
        df_new = reformat(df_new, action2)
        assert_frame_equal(df_new, df_expected)

    def test_reformat_currency(self):
        from mage_ai.data_cleaner.transformer_actions.column import reformat

        df_currency = pd.DataFrame(
            [
                ['$', '$    10000', 'stock exchange america', '$:MAGE', 5.34],
                ['£', '£200', 'huddersfield stock exchange', '£:XYZA', -1.34],
                ['CAD', 'CAD 100', None, '', -0.89],
                ['¥', '', 'stock exchange japan', '', 4.23],
                ['€', '€ 123.34', 'dresden stock exchange', '€:1234', 2.34],
                ['₹', '₹        10000', np.nan, '₹:FDSA', -7.80],
                ['Rs', 'Rs 10000', '', '₹:ASDF', 4.44],
                ['', '元10000', 'stock exchange china', '元:ASDF', 1.02],
                [None, None, 'stock exchange san jose', None, -2.01],
            ],
            columns=[
                'native_currency',
                'value',
                'exchange',
                'ticker',
                'growth_rate',
            ],
        )
        df_expected = pd.DataFrame(
            [
                ['$', 10000, 'stock exchange america', '$:MAGE', 5.34],
                ['£', 200, 'huddersfield stock exchange', '£:XYZA', -1.34],
                ['CAD', 100, None, '', -0.89],
                ['¥', np.nan, 'stock exchange japan', '', 4.23],
                ['€', 123.34, 'dresden stock exchange', '€:1234', 2.34],
                ['₹', 10000, np.nan, '₹:FDSA', -7.80],
                ['Rs', 10000, '', '₹:ASDF', 4.44],
                ['', 10000, 'stock exchange china', '元:ASDF', 1.02],
                [None, np.nan, 'stock exchange san jose', None, -2.01],
            ],
            columns=[
                'native_currency',
                'value',
                'exchange',
                'ticker',
                'growth_rate',
            ],
        )
        action_currency = dict(
            action_type='reformat',
            action_arguments=['value'],
            axis='column',
            action_options={
                'reformat': 'currency_to_num',
            },
            action_variables={
                'value': {
                    'feature': {'column_type': 'number_with_decimals', 'uuid': 'value'},
                    'type': 'feature',
                }
            },
            action_code='',
            outputs=[],
        )
        df_new = reformat(df_currency, action_currency)
        assert_frame_equal(df_new, df_expected)

    def test_currency_conversion_test_all_formatting(self):
        from mage_ai.data_cleaner.transformer_actions.column import reformat

        values = [
            '  $ 10000',
            '- ¥ 22.324523',
            'Rs 100000.23   ',
            '  € 12.23425',
            'CAD     12423      ',
            '£ .0927503',
            '-₹ 0',
            ' 10000 元   ',
            ' 0.42 €',
            ' -  3.42032 CAD',
        ]
        expected_values = [
            10000,
            -22.324523,
            100000.23,
            12.23425,
            12423,
            0.0927503,
            -0.0,
            10000,
            0.42,
            -3.42032,
        ]

        df = pd.DataFrame({'column': values})
        expected_df = pd.DataFrame({'column': expected_values})
        action = dict(
            action_type='reformat',
            action_arguments=['column'],
            axis='column',
            action_options={
                'reformat': 'currency_to_num',
            },
            action_variables={
                'column': {
                    'feature': {'column_type': 'number_with_decimals', 'uuid': 'column'},
                    'type': 'feature',
                }
            },
            action_code='',
            outputs=[],
        )
        new_df = reformat(df, action)
        assert_frame_equal(new_df, expected_df)

    def test_reformat_time(self):
        from mage_ai.data_cleaner.transformer_actions.column import reformat

        df = pd.DataFrame(
            [
                [dt(2022, 8, 4), None, 'Action Movie #1', 'not a date', 234],
                [dt(2022, 1, 20), '', 'sportsball', '1-20-2022', 13234],
                [None, '12/24/22', 'reality tv show', '12-24-2022', 23234],
                [dt(2022, 10, 31), '10/31/22', '', '10.31.2022', 21432],
                [dt(2022, 6, 27), None, 'action Movie #2', '6/27/2022', 324212],
                [dt(2022, 3, 8), '03/08/    22', 'game show', '3/8/2022', 2034],
            ],
            columns=[
                'date1',
                'date2',
                'notdate',
                'mostlydate',
                'date5',
            ],
        )
        expected_df = pd.DataFrame(
            [
                [dt(2022, 8, 4), pd.NaT, 'Action Movie #1', pd.NaT, pd.to_datetime(234)],
                [
                    dt(2022, 1, 20),
                    pd.NaT,
                    'sportsball',
                    pd.to_datetime('1-20-2022'),
                    pd.to_datetime(13234),
                ],
                [
                    None,
                    pd.to_datetime('12/24/22'),
                    'reality tv show',
                    pd.to_datetime('12-24-2022'),
                    pd.to_datetime(23234),
                ],
                [
                    dt(2022, 10, 31),
                    pd.to_datetime('10/31/22'),
                    '',
                    pd.to_datetime('10.31.2022'),
                    pd.to_datetime(21432),
                ],
                [
                    dt(2022, 6, 27),
                    pd.NaT,
                    'action Movie #2',
                    pd.to_datetime('6/27/2022'),
                    pd.to_datetime(324212),
                ],
                [
                    dt(2022, 3, 8),
                    pd.to_datetime('03/08/22'),
                    'game show',
                    pd.to_datetime('3/8/2022'),
                    pd.to_datetime(2034),
                ],
            ],
            columns=[
                'date1',
                'date2',
                'notdate',
                'mostlydate',
                'date5',
            ],
        )
        action = dict(
            action_type='reformat',
            action_arguments=['date2', 'mostlydate', 'date5'],
            axis='column',
            action_options={
                'reformat': 'date_format_conversion',
            },
            action_variables={
                'date2': {
                    'feature': {'column_type': 'datetime', 'uuid': 'date2'},
                    'type': 'feature',
                },
                'mostlydate': {
                    'feature': {'column_type': 'category_high_cardinality', 'uuid': 'mostlydate'},
                    'type': 'feature',
                },
                'date5': {'feature': {'column_type': 'number', 'uuid': 'date5'}, 'type': 'feature'},
            },
            action_code='',
            outputs=[],
        )
        df_new = reformat(df, action).reset_index(drop=True)
        assert_frame_equal(df_new, expected_df)

    def test_reformat_time_bad_inputs(self):
        from mage_ai.data_cleaner.transformer_actions.column import reformat

        df = pd.DataFrame(
            [
                [dt(2022, 8, 4), '08/04/22', 'Action Movie #1', 'not a date', np.nan, True],
                [dt(2022, 1, 20), '', 'sportsball', '1-20-2022', -1323.4, False],
                [None, '12/24/22', 'reality tv show', '12-24-2022', -232342.322, False],
                [dt(2022, 10, 31), '10/31/22', '', '10.31.2022', 21.432, None],
                [dt(2022, 6, 27), None, 'action Movie #2', '6/27/2022', 324212, True],
                [dt(2022, 3, 8), '03/08/    22', 'game show', '3/8/2022', -9830, False],
            ],
            columns=['date1', 'date2', 'notdate', 'mostlydate', 'date5', 'boolean'],
        )
        expected_df = pd.DataFrame(
            [
                [dt(2022, 8, 4), pd.to_datetime('08/04/22'), pd.NaT, pd.NaT, np.nan, pd.NaT],
                [
                    dt(2022, 1, 20),
                    pd.NaT,
                    pd.NaT,
                    pd.to_datetime('1-20-2022'),
                    pd.to_datetime(-1323.4),
                    pd.NaT,
                ],
                [
                    None,
                    pd.to_datetime('12/24/22'),
                    pd.NaT,
                    pd.to_datetime('12-24-2022'),
                    pd.to_datetime(-232342.322),
                    pd.NaT,
                ],
                [
                    dt(2022, 10, 31),
                    pd.to_datetime('10/31/22'),
                    pd.NaT,
                    pd.to_datetime('10.31.2022'),
                    pd.to_datetime(21.432),
                    pd.NaT,
                ],
                [
                    dt(2022, 6, 27),
                    pd.NaT,
                    pd.NaT,
                    pd.to_datetime('6/27/2022'),
                    pd.to_datetime(324212),
                    pd.NaT,
                ],
                [
                    dt(2022, 3, 8),
                    pd.to_datetime('03/08/22'),
                    pd.NaT,
                    pd.to_datetime('3/8/2022'),
                    pd.to_datetime(-9830),
                    pd.NaT,
                ],
            ],
            columns=['date1', 'date2', 'notdate', 'mostlydate', 'date5', 'boolean'],
        )
        action = dict(
            action_type='reformat',
            action_arguments=['date1', 'date2', 'notdate', 'mostlydate', 'date5', 'boolean'],
            axis='column',
            action_options={
                'reformat': 'date_format_conversion',
            },
            action_variables={
                'date1': {
                    'feature': {'column_type': 'datetime', 'uuid': 'date1'},
                    'type': 'feature',
                },
                'date2': {
                    'feature': {'column_type': 'datetime', 'uuid': 'date2'},
                    'type': 'feature',
                },
                'notdate': {
                    'feature': {'column_type': 'text', 'uuid': 'notdate'},
                    'type': 'feature',
                },
                'mostlydate': {
                    'feature': {'column_type': 'category_high_cardinality', 'uuid': 'mostlydate'},
                    'type': 'feature',
                },
                'date5': {
                    'feature': {'column_type': 'number_with_decimals', 'uuid': 'date5'},
                    'type': 'feature',
                },
                'boolean': {
                    'feature': {'column_type': 'true_or_false', 'uuid': 'boolean'},
                    'type': 'feature',
                },
            },
            action_code='',
            outputs=[],
        )
        df_new = reformat(df, action).reset_index(drop=True)
        assert_frame_equal(df_new, expected_df)

    def test_reformat_bad_inputs(self):
        from mage_ai.data_cleaner.transformer_actions.column import reformat

        df = pd.DataFrame(
            [
                [dt(2022, 8, 4), None, 'Action Movie #1', 'not a date', 234],
                [dt(2022, 1, 20), '', 'sportsball', '1-20-2022', 13234],
                [None, '12/24/22', 'reality tv show', '12-24-2022', 23234],
                [dt(2022, 10, 31), '10/31/22', '', '10.31.2022', 21432],
                [dt(2022, 6, 27), None, 'action Movie #2', '6/27/2022', 324212],
                [dt(2022, 3, 8), '03/08/    22', 'game show', '3/8/2022', 2034],
            ],
            columns=[
                'date1',
                'date2',
                'notdate',
                'mostlydate',
                'date5',
            ],
        )

        expected_df = pd.DataFrame(
            [
                [dt(2022, 8, 4), None, pd.NaT, pd.NaT, 234],
                [dt(2022, 1, 20), '', pd.NaT, pd.Timestamp('1-20-2022'), 13234],
                [None, '12/24/22', pd.NaT, pd.Timestamp('12-24-2022'), 23234],
                [dt(2022, 10, 31), '10/31/22', pd.NaT, pd.Timestamp('10.31.2022'), 21432],
                [dt(2022, 6, 27), None, pd.NaT, pd.Timestamp('6/27/2022'), 324212],
                [dt(2022, 3, 8), '03/08/    22', pd.NaT, pd.Timestamp('3/8/2022'), 2034],
            ],
            columns=[
                'date1',
                'date2',
                'notdate',
                'mostlydate',
                'date5',
            ],
        )

        action = dict(
            action_type='reformat',
            action_arguments=['date1', 'date5'],
            axis='column',
            action_options={'reformat': 'caps_standardization', 'capitalization': 'lowercase'},
            action_variables={},
            action_code='',
            outputs=[],
        )
        action2 = dict(
            action_type='reformat',
            action_arguments=['date1', 'date5', 'notdate', 'mostlydate'],
            axis='column',
            action_options={
                'reformat': 'currency_conversion',
            },
            action_variables={},
            action_code='',
            outputs=[],
        )
        action3 = dict(
            action_type='reformat',
            action_arguments=['notdate', 'mostlydate'],
            axis='column',
            action_options={
                'reformat': 'date_format_conversion',
            },
            action_variables={},
            action_code='',
            outputs=[],
        )
        df_new = reformat(df, action).reset_index(drop=True)
        assert_frame_equal(df_new, df)
        df_new2 = reformat(df, action2).reset_index(drop=True)
        assert_frame_equal(df_new2, df)
        df_new3 = reformat(df, action3).reset_index(drop=True)
        df_new3['notdate'] = df_new3['notdate'].astype(np.datetime64)
        df_new3['mostlydate'] = df_new3['mostlydate'].astype(np.datetime64)
        assert_frame_equal(df_new3, expected_df)

    def test_remove_outliers(self):
        from mage_ai.data_cleaner.transformer_actions.column import remove_outliers

        df = pd.DataFrame(
            [
                [1, 1, 1],
                [1, 2, 997],
                [1, 3, 998],
                [1, 4, 999],
                [1, 5, 1000],
                [1, 6, 997],
                [1, 7, 998],
                [1, 8, 999],
                [1, 9, 1000],
                [1, 10, 997],
                [1, 11, 998],
                [1, 12, 999],
            ],
            columns=['number1', 'number2', 'number3'],
        )

        expected_df = df.drop(0, axis=0)

        action_lof = dict(
            action_type='remove_outliers',
            action_arguments=['number1', 'number2', 'number3'],
            action_options={'method': 'lof'},
            action_variables={
                'number1': {
                    'feature': {'column_type': 'number', 'uuid': 'number1'},
                    'type': 'feature',
                },
                'number2': {
                    'feature': {'column_type': 'number', 'uuid': 'number2'},
                    'type': 'feature',
                },
                'number3': {
                    'feature': {'column_type': 'number', 'uuid': 'number3'},
                    'type': 'feature',
                },
            },
            axis='column',
        )

        action_itree = dict(
            action_type='remove_outliers',
            action_arguments=['number1', 'number2', 'number3'],
            action_options={'method': 'itree'},
            action_variables={
                'number1': {
                    'feature': {'column_type': 'number', 'uuid': 'number1'},
                    'type': 'feature',
                },
                'number2': {
                    'feature': {'column_type': 'number', 'uuid': 'number2'},
                    'type': 'feature',
                },
                'number3': {
                    'feature': {'column_type': 'number', 'uuid': 'number3'},
                    'type': 'feature',
                },
            },
            axis='column',
        )

        lof_df = remove_outliers(df, action_lof)
        itree_df = remove_outliers(df, action_itree)

        assert_frame_equal(expected_df, lof_df)
        assert_frame_equal(expected_df, itree_df)

    def test_remove_outliers_bad_inputs(self):
        from mage_ai.data_cleaner.transformer_actions.column import remove_outliers

        df = pd.DataFrame(
            [
                ['e', '2', 997],
                ['e', '3', np.nan],
                ['e', '4', 999],
                ['e', '5', 1000],
                ['e', '6', 997],
                ['e', None, 998],
                ['e', '8', 999],
                ['e', '9', 1000],
                ['e', '10', 997],
                ['e', '11', np.nan],
                ['e', '12', 999],
                ['e', '1', 1],
            ],
            columns=['number1', 'number2', 'number3'],
        )

        expected_df = df.drop(11, axis=0)

        action_one = dict(
            action_type='remove_outliers',
            action_arguments=['number2', 'number3'],
            action_options={'method': 'auto'},
            action_variables={
                'number1': {
                    'feature': {'column_type': 'text', 'uuid': 'number1'},
                    'type': 'feature',
                },
                'number2': {
                    'feature': {'column_type': 'number', 'uuid': 'number2'},
                    'type': 'feature',
                },
                'number3': {
                    'feature': {'column_type': 'number', 'uuid': 'number3'},
                    'type': 'feature',
                },
            },
            axis='column',
        )

        action_two = dict(
            action_type='remove_outliers',
            action_arguments=['number1', 'number3'],
            action_options={'method': 'auto'},
            action_variables={
                'number1': {
                    'feature': {'column_type': 'text', 'uuid': 'number1'},
                    'type': 'feature',
                },
                'number2': {
                    'feature': {'column_type': 'number', 'uuid': 'number2'},
                    'type': 'feature',
                },
                'number3': {
                    'feature': {'column_type': 'number', 'uuid': 'number3'},
                    'type': 'feature',
                },
            },
            axis='column',
        )

        df_one = remove_outliers(df, action_one)
        df_two = remove_outliers(df, action_two)

        assert_frame_equal(expected_df, df_one)
        assert_frame_equal(expected_df, df_two)

    def test_remove_outliers_lof(self):
        from mage_ai.data_cleaner.transformer_actions.column import remove_outliers

        """
        This test case is engineered so that only in the
        multidimensional case is a row removed as an outlier
        """
        df = pd.DataFrame(
            [
                [0, 0, -1],
                [0, 0, -2],
                [0, 0, -3],
                [0, 0, 2],
                [0, 0, 1],
                [0, 0, 3],
                [0, 0, 2],
                [0, 2, 0],
                [3, 0, 0],
                [1, 0, 0],
                [2, 0, 0],
                [3, 0, 0],
                [-2, 0, 0],
                [0, -1, 0],
                [0, 0, -2],
                [0, 2, 0],
                [0, 3, 0],
                [0, 4, 0],
                [0, 3, 0],
                [0, 2, 0],
                [-3, -2, -3],
                [0, -3, 0],
                [0, -2, 0],
                [0, -3, 0],
                [2, 2, 2],
                [0, -1, 0],
                [-1, 0, 0],
                [-3, 0, 0],
                [2, 0, 0],
                [-2, 0, 0],
            ],
            columns=['number1', 'number2', 'number3'],
        )

        expected_df = df.drop([20, 24], axis=0)

        action = dict(
            action_type='remove_outliers',
            action_arguments=['number1', 'number2', 'number3'],
            action_options={'method': 'auto'},
            action_variables={
                'number1': {
                    'feature': {'column_type': 'number', 'uuid': 'number1'},
                    'type': 'feature',
                },
                'number2': {
                    'feature': {'column_type': 'number', 'uuid': 'number2'},
                    'type': 'feature',
                },
                'number3': {
                    'feature': {'column_type': 'number', 'uuid': 'number3'},
                    'type': 'feature',
                },
            },
            axis='column',
        )

        new_df = remove_outliers(df, action)

        assert_frame_equal(new_df, expected_df)

    def test_remove_outliers_itree(self):
        from mage_ai.data_cleaner.transformer_actions.column import remove_outliers

        """
        This test case is engineered so that only in the
        multidimensional case is a row removed as an outlier
        """
        data = np.zeros((30, 7))
        col = np.random.randint(0, 6, size=(30,))
        values = np.random.randint(-3, 3, size=(30,))
        data[np.arange(30), col] = values
        data[0] = np.array([-10, 2, 4, -5, 7, 2, 3])
        data[7] = np.array([7, 3, -5, 4, 3, -9, -8])
        data[18] = np.array([-4, 5, 10, -3, -2, 2, -6])
        data[27] = np.array([-10, 3, 3, 4, 5, -9, 3])
        df = pd.DataFrame(data, columns=['1', '2', '3', '4', '5', '6', '7'])
        expected_df = df.drop([0, 7, 18, 27], axis=0)
        action = dict(
            action_type='remove_outliers',
            action_arguments=['1', '2', '3', '4', '5', '6', '7'],
            action_options={'method': 'auto'},
            action_variables={
                '1': {
                    'feature': {'column_type': 'number', 'uuid': '1'},
                    'type': 'feature',
                },
                '2': {
                    'feature': {'column_type': 'number', 'uuid': '2'},
                    'type': 'feature',
                },
                '3': {
                    'feature': {'column_type': 'number', 'uuid': '3'},
                    'type': 'feature',
                },
                '4': {
                    'feature': {'column_type': 'number', 'uuid': '4'},
                    'type': 'feature',
                },
                '5': {
                    'feature': {'column_type': 'number', 'uuid': '5'},
                    'type': 'feature',
                },
                '6': {
                    'feature': {'column_type': 'number', 'uuid': '6'},
                    'type': 'feature',
                },
                '7': {
                    'feature': {'column_type': 'number', 'uuid': '7'},
                    'type': 'feature',
                },
            },
            axis='column',
        )
        new_df = remove_outliers(df, action)
        assert_frame_equal(new_df, expected_df)

    def test_select(self):
        df = pd.DataFrame(
            [
                [1, 1000],
                [2, 1050],
            ],
            columns=[
                'group_id',
                'order_id',
            ],
        )
        action = dict(action_arguments=['group_id'])
        df_new = select(df, action)
        self.assertEqual(
            df_new.to_dict(orient='records'),
            [
                dict(
                    group_id=1,
                ),
                dict(
                    group_id=2,
                ),
            ],
        )

    def test_shift_down(self):
        df = pd.DataFrame(
            [
                ['2020-01-01', 1000],
                ['2020-01-02', 1050],
                ['2020-01-03', 1200],
                ['2020-01-04', 990],
            ],
            columns=[
                'date',
                'sold',
            ],
        )
        action = dict(
            action_arguments=['sold'],
            outputs=[
                dict(uuid='prev_sold'),
            ],
        )
        df_new = shift_down(df, action)
        self.assertEqual(
            df_new.to_dict(orient='records')[1:],
            [
                dict(
                    date='2020-01-02',
                    sold=1050,
                    prev_sold=1000,
                ),
                dict(
                    date='2020-01-03',
                    sold=1200,
                    prev_sold=1050,
                ),
                dict(
                    date='2020-01-04',
                    sold=990,
                    prev_sold=1200,
                ),
            ],
        )

    def test_shift_down_with_groupby(self):
        df = pd.DataFrame(
            [
                [1, '2020-01-01', 1000],
                [1, '2020-01-02', 1050],
                [2, '2020-01-03', 1200],
                [1, '2020-01-04', 990],
                [2, '2020-01-05', 980],
                [2, '2020-01-06', 970],
                [2, '2020-01-07', 960],
            ],
            columns=[
                'group_id',
                'date',
                'sold',
            ],
        )
        action = dict(
            action_arguments=['sold'],
            action_options=dict(
                groupby_columns=['group_id'],
                periods=2,
            ),
            outputs=[
                dict(uuid='prev_sold'),
            ],
        )
        df_new = shift_down(df, action)
        df_expected = pd.DataFrame(
            [
                [1, '2020-01-01', 1000, None],
                [1, '2020-01-02', 1050, None],
                [2, '2020-01-03', 1200, None],
                [1, '2020-01-04', 990, 1000],
                [2, '2020-01-05', 980, None],
                [2, '2020-01-06', 970, 1200],
                [2, '2020-01-07', 960, 980],
            ],
            columns=[
                'group_id',
                'date',
                'sold',
                'prev_sold',
            ],
        )
        assert_frame_equal(df_new, df_expected)

    def test_shift_up(self):
        df = pd.DataFrame(
            [
                ['2020-01-01', 1000],
                ['2020-01-02', 1050],
                ['2020-01-03', 1200],
                ['2020-01-04', 990],
            ],
            columns=[
                'date',
                'sold',
            ],
        )
        action = dict(
            action_arguments=['sold'],
            outputs=[
                dict(uuid='next_sold'),
            ],
        )
        df_new = shift_up(df, action)
        self.assertEqual(
            df_new.to_dict(orient='records')[:-1],
            [
                dict(
                    date='2020-01-01',
                    sold=1000,
                    next_sold=1050,
                ),
                dict(
                    date='2020-01-02',
                    sold=1050,
                    next_sold=1200,
                ),
                dict(
                    date='2020-01-03',
                    sold=1200,
                    next_sold=990,
                ),
            ],
        )

    def test_sum(self):
        from mage_ai.data_cleaner.transformer_actions.column import sum

        action = self.__groupby_agg_action('total_amount')
        df_new = sum(TEST_DATAFRAME.copy(), action)
        self.assertEqual(
            df_new.to_dict(orient='records'),
            [
                dict(
                    group_id=1,
                    amount=1000,
                    total_amount=2100,
                ),
                dict(
                    group_id=2,
                    amount=1050,
                    total_amount=2200,
                ),
                dict(
                    group_id=1,
                    amount=1100,
                    total_amount=2100,
                ),
                dict(
                    group_id=2,
                    amount=1150,
                    total_amount=2200,
                ),
            ],
        )

    def __groupby_agg_action(self, output_col):
        return dict(
            action_arguments=['amount'],
            action_options=dict(groupby_columns=['group_id']),
            outputs=[
                dict(uuid=output_col),
            ],
        )



"""Add secrets table

Revision ID: 1f9353eddbc6
Revises: 643b6e65e814
Create Date: 2023-02-15 09:25:39.658782

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '1f9353eddbc6'
down_revision = '643b6e65e814'
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('secret',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('created_at', sa.DateTime(timezone=True), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=True),
    sa.Column('updated_at', sa.DateTime(timezone=True), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=True),
    sa.Column('name', sa.String(length=255), nullable=True),
    sa.Column('value', sa.Text(), nullable=True),
    sa.Column('repo_name', sa.String(length=255), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('name')
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('secret')
    # ### end Alembic commands ###
